#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : gen_action_details

import random

from metagpt.environment.stanford_town.env_space import EnvObsParams, EnvObsType
from metagpt.ext.stanford_town.actions.st_action import STAction
from metagpt.logs import logger


class GenActionSector(STAction):
    name: str = "GenActionSector"

    def _func_cleanup(self, llm_resp: str, prompt: str):
        cleaned_response = llm_resp.split("}")[0]
        return cleaned_response

    def _func_validate(self, llm_resp: str, prompt: str):
        if len(llm_resp.strip()) < 1:
            return False
        if "}" not in llm_resp:
            return False
        if "," in llm_resp:
            return False
        return True

    def _func_fail_default_resp(self):
        fs = "kitchen"
        return fs

    async def run(self, role: "STRole", access_tile: dict[str, str], act_desp: str):
        def create_prompt_input(role, access_tile: dict[str, str], act_desp):
            act_world = f"{access_tile['world']}"

            prompt_input = []

            prompt_input += [role.scratch.get_str_name()]
            prompt_input += [role.scratch.living_area.split(":")[1]]
            x = f"{act_world}:{role.scratch.living_area.split(':')[1]}"
            prompt_input += [role.s_mem.get_str_accessible_sector_arenas(x)]

            prompt_input += [role.scratch.get_str_name()]
            prompt_input += [f"{access_tile['sector']}"]
            x = f"{act_world}:{access_tile['sector']}"
            prompt_input += [role.s_mem.get_str_accessible_sector_arenas(x)]

            if role.scratch.get_str_daily_plan_req() != "":
                prompt_input += [f"\n{role.scratch.get_str_daily_plan_req()}"]
            else:
                prompt_input += [""]

            # MAR 11 TEMP
            prompt_input = []
            act_world = access_tile["world"]
            accessible_sector_str = role.s_mem.get_str_accessible_sectors(act_world)
            curr = accessible_sector_str.split(", ")
            fin_accessible_sectors = []
            for i in curr:
                if "'s house" in i:
                    if role.scratch.last_name in i:
                        fin_accessible_sectors += [i]
            else:
                fin_accessible_sectors += [i]
            accessible_sector_str = ", ".join(fin_accessible_sectors)
            # END MAR 11 TEMP

            prompt_input += [accessible_sector_str]

            act_desp_1 = act_desp
            act_desp_2 = act_desp
            if "(" in act_desp:
                act_desp_1 = act_desp.split("(")[0].strip()
                act_desp_2 = act_desp.split("(")[-1][:-1]
            prompt_input += [role.scratch.get_str_name()]
            prompt_input += [act_desp_1]

            prompt_input += [act_desp_2]
            prompt_input += [role.scratch.get_str_name()]
            return prompt_input

        prompt_template = "action_location_sector_v1.txt"
        prompt_input = create_prompt_input(role, access_tile, act_desp)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, prompt_template)

        self.fail_default_resp = self._func_fail_default_resp()
        output = await self._run_gpt35_max_tokens(prompt, max_tokens=15)
        y = f"{access_tile['world']}"
        x = [i.strip() for i in role.s_mem.get_str_accessible_sectors(y).split(",")]
        if output not in x:
            # output = random.choice(x)
            output = role.scratch.living_area.split(":")[1]
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output


class GenActionArena(STAction):
    name: str = "GenActionArena"

    def _func_cleanup(self, llm_resp: str, prompt: str):
        cleaned_response = llm_resp.split("}")[0]
        return cleaned_response

    def _func_validate(self, llm_resp: str, prompt: str):
        if len(llm_resp.strip()) < 1:
            return False
        if "}" not in llm_resp:
            return False
        if "," in llm_resp:
            return False
        return True

    def _func_fail_default_resp(self):
        fs = "kitchen"
        return fs

    async def run(self, role: "STRole", act_desp: str, act_world: str, act_sector: str):
        def create_prompt_input(role, act_desp, act_world, act_sector):
            prompt_input = []
            prompt_input += [role.scratch.get_str_name()]
            x = f"{act_world}:{act_sector}"
            prompt_input += [act_sector]

            # MAR 11 TEMP
            accessible_arena_str = role.s_mem.get_str_accessible_sector_arenas(x)
            curr = accessible_arena_str.split(", ")
            fin_accessible_arenas = []
            for i in curr:
                if "'s room" in i:
                    if role.scratch.last_name in i:
                        fin_accessible_arenas += [i]
                else:
                    fin_accessible_arenas += [i]
            accessible_arena_str = ", ".join(fin_accessible_arenas)
            # END MAR 11 TEMP
            prompt_input += [accessible_arena_str]
            act_desp_1 = act_desp
            act_desp_2 = act_desp
            if "(" in act_desp:
                act_desp_1 = act_desp.split("(")[0].strip()
                act_desp_2 = act_desp.split("(")[-1][:-1]
            prompt_input += [role.scratch.get_str_name()]
            prompt_input += [act_desp_1]

            prompt_input += [act_desp_2]
            prompt_input += [role.scratch.get_str_name()]

            prompt_input += [act_sector]
            prompt_input += [accessible_arena_str]
            return prompt_input

        prompt_template = "action_location_object_vMar11.txt"
        prompt_input = create_prompt_input(role, act_desp, act_world, act_sector)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, prompt_template)
        self.fail_default_resp = self._func_fail_default_resp()
        output = await self._run_gpt35_max_tokens(prompt, max_tokens=15)
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output


class GenActionObject(STAction):
    name: str = "GenActionObject"

    def _func_validate(self, llm_resp: str, prompt: str):
        if len(llm_resp.strip()) < 1:
            return False
        return True

    def _func_cleanup(self, llm_resp: str, prompt: str):
        cleaned_response = llm_resp.strip()
        return cleaned_response

    def _func_fail_default_resp(self):
        fs = "bed"
        return fs

    async def run(self, role: "STRole", act_desp: str, temp_address: str):
        def create_prompt_input(role, act_desp, temp_address):
            prompt_input = []
            if "(" in act_desp:
                act_desp = act_desp.split("(")[-1][:-1]

            prompt_input += [act_desp]
            prompt_input += [role.s_mem.get_str_accessible_arena_game_objects(temp_address)]
            return prompt_input

        prompt_template = "action_object_v2.txt"
        prompt_input = create_prompt_input(role, act_desp, temp_address)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, prompt_template)
        self.fail_default_resp = self._func_fail_default_resp()
        output = await self._run_gpt35_max_tokens(prompt, max_tokens=15)
        x = [i.strip() for i in role.s_mem.get_str_accessible_arena_game_objects(temp_address).split(",")]
        if output not in x:
            output = random.choice(x)
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output


class GenPronunciatio(STAction):
    name: str = "GenPronunciatio"

    def _func_cleanup(self, llm_resp: str, prompt: str):
        cr = llm_resp.strip()
        if len(cr) > 3:
            cr = cr[:3]
        return cr

    def _func_validate(self, llm_resp: str, prompt: str):
        try:
            self._func_cleanup(llm_resp, prompt="")
            if len(llm_resp) == 0:
                return False
        except Exception:
            return False
        return True

    def _func_fail_default_resp(self):
        fs = "😋"
        return fs

    async def run(self, role: "STRole", act_desp: str):
        def create_prompt_input(act_desp):
            if "(" in act_desp:
                act_desp = act_desp.split("(")[-1].split(")")[0]
            prompt_input = [act_desp]
            return prompt_input

        prompt_template = "generate_pronunciatio_v1.txt"
        prompt_input = create_prompt_input(act_desp)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, prompt_template)
        example_output = "🛁🧖‍♀️"
        special_instruction = "The value for the output must ONLY contain the emojis."
        self.fail_default_resp = self._func_fail_default_resp()
        output = await self._run_gpt35(prompt, example_output, special_instruction)
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output


class GenEventTriple(STAction):
    name: str = "GenEventTriple"

    def _func_cleanup(self, llm_resp: str, prompt: str):
        cr = llm_resp.strip()
        cr = [i.strip() for i in cr.split(")")[0].split(",")]
        return cr

    def _func_validate(self, llm_resp: str, prompt: str):
        try:
            llm_resp = self._func_cleanup(llm_resp, prompt="")
            if len(llm_resp) != 2:
                return False
        except Exception:
            return False
        return True

    def _func_fail_default_resp(self, role):
        fs = (role.name, "is", "idle")
        return fs

    async def run(self, role: "STRole", act_desp: str):
        def create_prompt_input(role, act_desp):
            if "(" in act_desp:
                act_desp = act_desp.split("(")[-1].split(")")[0]
            prompt_input = [role.name, act_desp, role.name]
            return prompt_input

        prompt_template = "generate_event_triple_v1.txt"
        prompt_input = create_prompt_input(role, act_desp)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, prompt_template)
        self.fail_default_resp = self._func_fail_default_resp(role)
        output = await self._run_gpt35_max_tokens(prompt, max_tokens=30)
        output = (role.name, output[0], output[1])
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output


class GenActObjDescription(STAction):
    name: str = "GenActObjDescription"

    def _func_cleanup(self, llm_resp: str, prompt: str):
        cr = llm_resp.strip()
        if cr[-1] == ".":
            cr = cr[:-1]
        return cr

    def _func_validate(self, llm_resp: str, prompt: str):
        try:
            llm_resp = self._func_cleanup(llm_resp, prompt="")
        except Exception:
            return False
        return True

    def _func_fail_default_resp(self, act_game_object):
        fs = f"{act_game_object} is idle"
        return fs

    async def run(self, role: "STRole", act_game_object: str, act_desp: str):
        def create_prompt_input(act_game_object, act_desp, role):
            prompt_input = [act_game_object, role.name, act_desp, act_game_object, act_game_object]
            return prompt_input

        prompt_template = "generate_obj_event_v1.txt"
        prompt_input = create_prompt_input(act_game_object, act_desp, role)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, prompt_template)
        example_output = "being fixed"
        special_instruction = "The output should ONLY contain the phrase that should go in <fill in>."
        self.fail_default_resp = self._func_fail_default_resp(act_game_object)
        output = await self._run_gpt35(prompt, example_output, special_instruction)
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output


class GenObjEventTriple(STAction):
    name: str = "GenObjEventTriple"

    def _func_cleanup(self, llm_resp: str, prompt: str):
        cr = llm_resp.strip()
        cr = [i.strip() for i in cr.split(")")[0].split(",")]
        return cr

    def _func_validate(self, llm_resp: str, prompt: str):
        try:
            llm_resp = self._func_cleanup(llm_resp, prompt="")
            if len(llm_resp) != 2:
                return False
        except Exception:
            return False
        return True

    def _func_fail_default_resp(self, act_game_object: str):
        fs = (act_game_object, "is", "idle")
        return fs

    async def run(self, role: "STRole", act_game_object, act_obj_desp):
        def create_prompt_input(act_game_object, act_obj_desp):
            prompt_input = [act_game_object, act_obj_desp, act_game_object]
            return prompt_input

        prompt_template = "generate_event_triple_v1.txt"
        prompt_input = create_prompt_input(act_game_object, act_obj_desp)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, prompt_template)
        self.fail_default_resp = self._func_fail_default_resp(act_game_object)
        output = await self._run_gpt35_max_tokens(prompt, max_tokens=30)
        output = (act_game_object, output[0], output[1])
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output


class GenActionDetails(STAction):
    name: str = "GenActionDetails"

    def _func_cleanup(self, llm_resp: str, prompt: str) -> list:
        pass

    def _func_validate(self, llm_resp: str, prompt: str) -> bool:
        # TODO -- this sometimes generates error
        try:
            self._func_cleanup(llm_resp)
        except Exception:
            return False
        return True

    def _func_fail_default_resp(self):
        fs = {}
        return fs

    async def run(self, role: "STRole", act_desp: str, act_dura):
        access_tile = role.rc.env.observe(
            obs_params=EnvObsParams(obs_type=EnvObsType.GET_TITLE, coord=role.scratch.curr_tile)
        )
        act_world = access_tile["world"]
        act_sector = await GenActionSector().run(role, access_tile, act_desp)
        act_arena = await GenActionArena().run(role, act_desp, act_world, act_sector)
        act_address = f"{act_world}:{act_sector}:{act_arena}"
        if not role.s_mem.get_str_accessible_arena_game_objects(act_address):
            act_game_object = "<random>"
        else:
            act_game_object = await GenActionObject().run(role, act_desp, act_address)
        new_address = f"{act_world}:{act_sector}:{act_arena}:{act_game_object}"
        act_pron = await GenPronunciatio().run(role, act_desp)
        act_event = await GenEventTriple().run(role, act_desp)
        # Persona's actions also influence the object states. We set those up here.
        act_obj_desp = await GenActObjDescription().run(role, act_game_object, act_desp)
        act_obj_pron = await GenPronunciatio().run(role, act_obj_desp)
        act_obj_event = await GenObjEventTriple().run(role, act_game_object, act_obj_desp)
        result_dict = {
            "action_address": new_address,
            "action_duration": int(act_dura),
            "action_description": act_desp,
            "action_pronunciatio": act_pron,
            "action_event": act_event,
            "chatting_with": None,
            "chat": None,
            "chatting_with_buffer": None,
            "chatting_end_time": None,
            "act_obj_description": act_obj_desp,
            "act_obj_pronunciatio": act_obj_pron,
            "act_obj_event": act_obj_event,
        }
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {result_dict}")
        return result_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Build a symbols repository from source code.

This script is designed to create a symbols repository from the provided source code.

@Time    : 2023/11/17 17:58
@Author  : alexanderwu
@File    : repo_parser.py
"""
from __future__ import annotations

import ast
import json
import re
import subprocess
from pathlib import Path
from typing import Dict, List, Optional

import pandas as pd
from pydantic import BaseModel, Field, field_validator

from metagpt.const import AGGREGATION, COMPOSITION, GENERALIZATION
from metagpt.logs import logger
from metagpt.utils.common import any_to_str, aread, remove_white_spaces
from metagpt.utils.exceptions import handle_exception


class RepoFileInfo(BaseModel):
    """
    Repository data element that represents information about a file.

    Attributes:
        file (str): The name or path of the file.
        classes (List): A list of class names present in the file.
        functions (List): A list of function names present in the file.
        globals (List): A list of global variable names present in the file.
        page_info (List): A list of page-related information associated with the file.
    """

    file: str
    classes: List = Field(default_factory=list)
    functions: List = Field(default_factory=list)
    globals: List = Field(default_factory=list)
    page_info: List = Field(default_factory=list)


class CodeBlockInfo(BaseModel):
    """
    Repository data element representing information about a code block.

    Attributes:
        lineno (int): The starting line number of the code block.
        end_lineno (int): The ending line number of the code block.
        type_name (str): The type or category of the code block.
        tokens (List): A list of tokens present in the code block.
        properties (Dict): A dictionary containing additional properties associated with the code block.
    """

    lineno: int
    end_lineno: int
    type_name: str
    tokens: List = Field(default_factory=list)
    properties: Dict = Field(default_factory=dict)


class DotClassAttribute(BaseModel):
    """
    Repository data element representing a class attribute in dot format.

    Attributes:
        name (str): The name of the class attribute.
        type_ (str): The type of the class attribute.
        default_ (str): The default value of the class attribute.
        description (str): A description of the class attribute.
        compositions (List[str]): A list of compositions associated with the class attribute.
    """

    name: str = ""
    type_: str = ""
    default_: str = ""
    description: str
    compositions: List[str] = Field(default_factory=list)

    @classmethod
    def parse(cls, v: str) -> "DotClassAttribute":
        """
        Parses dot format text and returns a DotClassAttribute object.

        Args:
            v (str): Dot format text to be parsed.

        Returns:
            DotClassAttribute: An instance of the DotClassAttribute class representing the parsed data.
        """
        val = ""
        meet_colon = False
        meet_equals = False
        for c in v:
            if c == ":":
                meet_colon = True
            elif c == "=":
                meet_equals = True
                if not meet_colon:
                    val += ":"
                    meet_colon = True
            val += c
        if not meet_colon:
            val += ":"
        if not meet_equals:
            val += "="

        cix = val.find(":")
        eix = val.rfind("=")
        name = val[0:cix].strip()
        type_ = val[cix + 1 : eix]
        default_ = val[eix + 1 :].strip()

        type_ = remove_white_spaces(type_)  # remove white space
        if type_ == "NoneType":
            type_ = ""
        if "Literal[" in type_:
            pre_l, literal, post_l = cls._split_literal(type_)
            composition_val = pre_l + "Literal" + post_l  # replace Literal[...] with Literal
            type_ = pre_l + literal + post_l
        else:
            type_ = re.sub(r"['\"]+", "", type_)  # remove '"
            composition_val = type_

        if default_ == "None":
            default_ = ""
        compositions = cls.parse_compositions(composition_val)
        return cls(name=name, type_=type_, default_=default_, description=v, compositions=compositions)

    @staticmethod
    def parse_compositions(types_part) -> List[str]:
        """
        Parses the type definition code block of source code and returns a list of compositions.

        Args:
            types_part: The type definition code block to be parsed.

        Returns:
            List[str]: A list of compositions extracted from the type definition code block.
        """
        if not types_part:
            return []
        modified_string = re.sub(r"[\[\],\(\)]", "|", types_part)
        types = modified_string.split("|")
        filters = {
            "str",
            "frozenset",
            "set",
            "int",
            "float",
            "complex",
            "bool",
            "dict",
            "list",
            "Union",
            "Dict",
            "Set",
            "Tuple",
            "NoneType",
            "None",
            "Any",
            "Optional",
            "Iterator",
            "Literal",
            "List",
        }
        result = set()
        for t in types:
            t = re.sub(r"['\"]+", "", t.strip())
            if t and t not in filters:
                result.add(t)
        return list(result)

    @staticmethod
    def _split_literal(v):
        """
        Parses the literal definition code block and returns three parts: pre-part, literal-part, and post-part.

        Args:
            v: The literal definition code block to be parsed.

        Returns:
            Tuple[str, str, str]: A tuple containing the pre-part, literal-part, and post-part of the code block.
        """
        tag = "Literal["
        bix = v.find(tag)
        eix = len(v) - 1
        counter = 1
        for i in range(bix + len(tag), len(v) - 1):
            c = v[i]
            if c == "[":
                counter += 1
                continue
            if c == "]":
                counter -= 1
                if counter > 0:
                    continue
                eix = i
                break
        pre_l = v[0:bix]
        post_l = v[eix + 1 :]
        pre_l = re.sub(r"['\"]", "", pre_l)  # remove '"
        pos_l = re.sub(r"['\"]", "", post_l)  # remove '"

        return pre_l, v[bix : eix + 1], pos_l

    @field_validator("compositions", mode="after")
    @classmethod
    def sort(cls, lst: List) -> List:
        """
        Auto-sorts a list attribute after making changes.

        Args:
            lst (List): The list attribute to be sorted.

        Returns:
            List: The sorted list.
        """
        lst.sort()
        return lst


class DotClassInfo(BaseModel):
    """
    Repository data element representing information about a class in dot format.

    Attributes:
        name (str): The name of the class.
        package (Optional[str]): The package to which the class belongs (optional).
        attributes (Dict[str, DotClassAttribute]): A dictionary of attributes associated with the class.
        methods (Dict[str, DotClassMethod]): A dictionary of methods associated with the class.
        compositions (List[str]): A list of compositions associated with the class.
        aggregations (List[str]): A list of aggregations associated with the class.
    """

    name: str
    package: Optional[str] = None
    attributes: Dict[str, DotClassAttribute] = Field(default_factory=dict)
    methods: Dict[str, DotClassMethod] = Field(default_factory=dict)
    compositions: List[str] = Field(default_factory=list)
    aggregations: List[str] = Field(default_factory=list)

    @field_validator("compositions", "aggregations", mode="after")
    @classmethod
    def sort(cls, lst: List) -> List:
        """
        Auto-sorts a list attribute after making changes.

        Args:
            lst (List): The list attribute to be sorted.

        Returns:
            List: The sorted list.
        """
        lst.sort()
        return lst


class DotClassRelationship(BaseModel):
    """
    Repository data element representing a relationship between two classes in dot format.

    Attributes:
        src (str): The source class of the relationship.
        dest (str): The destination class of the relationship.
        relationship (str): The type or nature of the relationship.
        label (Optional[str]): An optional label associated with the relationship.
    """

    src: str = ""
    dest: str = ""
    relationship: str = ""
    label: Optional[str] = None


class DotReturn(BaseModel):
    """
    Repository data element representing a function or method return type in dot format.

    Attributes:
        type_ (str): The type of the return.
        description (str): A description of the return type.
        compositions (List[str]): A list of compositions associated with the return type.
    """

    type_: str = ""
    description: str
    compositions: List[str] = Field(default_factory=list)

    @classmethod
    def parse(cls, v: str) -> "DotReturn" | None:
        """
        Parses the return type part of dot format text and returns a DotReturn object.

        Args:
            v (str): The dot format text containing the return type part to be parsed.

        Returns:
            DotReturn | None: An instance of the DotReturn class representing the parsed return type,
                             or None if parsing fails.
        """
        if not v:
            return DotReturn(description=v)
        type_ = remove_white_spaces(v)
        compositions = DotClassAttribute.parse_compositions(type_)
        return cls(type_=type_, description=v, compositions=compositions)

    @field_validator("compositions", mode="after")
    @classmethod
    def sort(cls, lst: List) -> List:
        """
        Auto-sorts a list attribute after making changes.

        Args:
            lst (List): The list attribute to be sorted.

        Returns:
            List: The sorted list.
        """
        lst.sort()
        return lst


class DotClassMethod(BaseModel):
    name: str
    args: List[DotClassAttribute] = Field(default_factory=list)
    return_args: Optional[DotReturn] = None
    description: str
    aggregations: List[str] = Field(default_factory=list)

    @classmethod
    def parse(cls, v: str) -> "DotClassMethod":
        """
        Parses a dot format method text and returns a DotClassMethod object.

        Args:
            v (str): The dot format text containing method information to be parsed.

        Returns:
            DotClassMethod: An instance of the DotClassMethod class representing the parsed method.
        """
        bix = v.find("(")
        eix = v.rfind(")")
        rix = v.rfind(":")
        if rix < 0 or rix < eix:
            rix = eix
        name_part = v[0:bix].strip()
        args_part = v[bix + 1 : eix].strip()
        return_args_part = v[rix + 1 :].strip()

        name = cls._parse_name(name_part)
        args = cls._parse_args(args_part)
        return_args = DotReturn.parse(return_args_part)
        aggregations = set()
        for i in args:
            aggregations.update(set(i.compositions))
        aggregations.update(set(return_args.compositions))

        return cls(name=name, args=args, description=v, return_args=return_args, aggregations=list(aggregations))

    @staticmethod
    def _parse_name(v: str) -> str:
        """
        Parses the dot format method name part and returns the method name.

        Args:
            v (str): The dot format text containing the method name part to be parsed.

        Returns:
            str: The parsed method name.
        """
        tags = [">", "</"]
        if tags[0] in v:
            bix = v.find(tags[0]) + len(tags[0])
            eix = v.rfind(tags[1])
            return v[bix:eix].strip()
        return v.strip()

    @staticmethod
    def _parse_args(v: str) -> List[DotClassAttribute]:
        """
        Parses the dot format method arguments part and returns the parsed arguments.

        Args:
            v (str): The dot format text containing the arguments part to be parsed.

        Returns:
            str: The parsed method arguments.
        """
        if not v:
            return []
        parts = []
        bix = 0
        counter = 0
        for i in range(0, len(v)):
            c = v[i]
            if c == "[":
                counter += 1
                continue
            elif c == "]":
                counter -= 1
                continue
            elif c == "," and counter == 0:
                parts.append(v[bix:i].strip())
                bix = i + 1
        parts.append(v[bix:].strip())

        attrs = []
        for p in parts:
            if p:
                attr = DotClassAttribute.parse(p)
                attrs.append(attr)
        return attrs


class RepoParser(BaseModel):
    """
    Tool to build a symbols repository from a project directory.

    Attributes:
        base_directory (Path): The base directory of the project.
    """

    base_directory: Path = Field(default=None)

    @classmethod
    @handle_exception(exception_type=Exception, default_return=[])
    def _parse_file(cls, file_path: Path) -> list:
        """
        Parses a Python file in the repository.

        Args:
            file_path (Path): The path to the Python file to be parsed.

        Returns:
            list: A list containing the parsed symbols from the file.
        """
        return ast.parse(file_path.read_text()).body

    def extract_class_and_function_info(self, tree, file_path) -> RepoFileInfo:
        """
        Extracts class, function, and global variable information from the Abstract Syntax Tree (AST).

        Args:
            tree: The Abstract Syntax Tree (AST) of the Python file.
            file_path: The path to the Python file.

        Returns:
            RepoFileInfo: A RepoFileInfo object containing the extracted information.
        """
        file_info = RepoFileInfo(file=str(file_path.relative_to(self.base_directory)))
        for node in tree:
            info = RepoParser.node_to_str(node)
            if info:
                file_info.page_info.append(info)
            if isinstance(node, ast.ClassDef):
                class_methods = [m.name for m in node.body if is_func(m)]
                file_info.classes.append({"name": node.name, "methods": class_methods})
            elif is_func(node):
                file_info.functions.append(node.name)
            elif isinstance(node, (ast.Assign, ast.AnnAssign)):
                for target in node.targets if isinstance(node, ast.Assign) else [node.target]:
                    if isinstance(target, ast.Name):
                        file_info.globals.append(target.id)
        return file_info

    def generate_symbols(self) -> List[RepoFileInfo]:
        """
        Builds a symbol repository from '.py' and '.js' files in the project directory.

        Returns:
            List[RepoFileInfo]: A list of RepoFileInfo objects containing the extracted information.
        """
        files_classes = []
        directory = self.base_directory

        matching_files = []
        extensions = ["*.py"]
        for ext in extensions:
            matching_files += directory.rglob(ext)
        for path in matching_files:
            tree = self._parse_file(path)
            file_info = self.extract_class_and_function_info(tree, path)
            files_classes.append(file_info)

        return files_classes

    def generate_json_structure(self, output_path: Path):
        """
        Generates a JSON file documenting the repository structure.

        Args:
            output_path (Path): The path to the JSON file to be generated.
        """
        files_classes = [i.model_dump() for i in self.generate_symbols()]
        output_path.write_text(json.dumps(files_classes, indent=4))

    def generate_dataframe_structure(self, output_path: Path):
        """
        Generates a DataFrame documenting the repository structure and saves it as a CSV file.

        Args:
            output_path (Path): The path to the CSV file to be generated.
        """
        files_classes = [i.model_dump() for i in self.generate_symbols()]
        df = pd.DataFrame(files_classes)
        df.to_csv(output_path, index=False)

    def generate_structure(self, output_path: str | Path = None, mode="json") -> Path:
        """
        Generates the structure of the repository in a specified format.

        Args:
            output_path (str | Path): The path to the output file or directory. Default is None.
            mode (str): The output format mode. Options: "json" (default), "csv", etc.

        Returns:
            Path: The path to the generated output file or directory.
        """
        output_file = self.base_directory / f"{self.base_directory.name}-structure.{mode}"
        output_path = Path(output_path) if output_path else output_file

        if mode == "json":
            self.generate_json_structure(output_path)
        elif mode == "csv":
            self.generate_dataframe_structure(output_path)
        return output_path

    @staticmethod
    def node_to_str(node) -> CodeBlockInfo | None:
        """
        Parses and converts an Abstract Syntax Tree (AST) node to a CodeBlockInfo object.

        Args:
            node: The AST node to be converted.

        Returns:
            CodeBlockInfo | None: A CodeBlockInfo object representing the parsed AST node,
                                  or None if the conversion fails.
        """
        if isinstance(node, ast.Try):
            return None
        if any_to_str(node) == any_to_str(ast.Expr):
            return CodeBlockInfo(
                lineno=node.lineno,
                end_lineno=node.end_lineno,
                type_name=any_to_str(node),
                tokens=RepoParser._parse_expr(node),
            )
        mappings = {
            any_to_str(ast.Import): lambda x: [RepoParser._parse_name(n) for n in x.names],
            any_to_str(ast.Assign): RepoParser._parse_assign,
            any_to_str(ast.ClassDef): lambda x: x.name,
            any_to_str(ast.FunctionDef): lambda x: x.name,
            any_to_str(ast.ImportFrom): lambda x: {
                "module": x.module,
                "names": [RepoParser._parse_name(n) for n in x.names],
            },
            any_to_str(ast.If): RepoParser._parse_if,
            any_to_str(ast.AsyncFunctionDef): lambda x: x.name,
            any_to_str(ast.AnnAssign): lambda x: RepoParser._parse_variable(x.target),
        }
        func = mappings.get(any_to_str(node))
        if func:
            code_block = CodeBlockInfo(lineno=node.lineno, end_lineno=node.end_lineno, type_name=any_to_str(node))
            val = func(node)
            if isinstance(val, dict):
                code_block.properties = val
            elif isinstance(val, list):
                code_block.tokens = val
            elif isinstance(val, str):
                code_block.tokens = [val]
            else:
                raise NotImplementedError(f"Not implement:{val}")
            return code_block
        logger.warning(f"Unsupported code block:{node.lineno}, {node.end_lineno}, {any_to_str(node)}")
        return None

    @staticmethod
    def _parse_expr(node) -> List:
        """
        Parses an expression Abstract Syntax Tree (AST) node.

        Args:
            node: The AST node representing an expression.

        Returns:
            List: A list containing the parsed information from the expression node.
        """
        funcs = {
            any_to_str(ast.Constant): lambda x: [any_to_str(x.value), RepoParser._parse_variable(x.value)],
            any_to_str(ast.Call): lambda x: [any_to_str(x.value), RepoParser._parse_variable(x.value.func)],
            any_to_str(ast.Tuple): lambda x: [any_to_str(x.value), RepoParser._parse_variable(x.value)],
        }
        func = funcs.get(any_to_str(node.value))
        if func:
            return func(node)
        raise NotImplementedError(f"Not implement: {node.value}")

    @staticmethod
    def _parse_name(n):
        """
        Gets the 'name' value of an Abstract Syntax Tree (AST) node.

        Args:
            n: The AST node.

        Returns:
            The 'name' value of the AST node.
        """
        if n.asname:
            return f"{n.name} as {n.asname}"
        return n.name

    @staticmethod
    def _parse_if(n):
        """
        Parses an 'if' statement Abstract Syntax Tree (AST) node.

        Args:
            n: The AST node representing an 'if' statement.

        Returns:
            None or Parsed information from the 'if' statement node.
        """
        tokens = []
        try:
            if isinstance(n.test, ast.BoolOp):
                tokens = []
                for v in n.test.values:
                    tokens.extend(RepoParser._parse_if_compare(v))
                return tokens
            if isinstance(n.test, ast.Compare):
                v = RepoParser._parse_variable(n.test.left)
                if v:
                    tokens.append(v)
            if isinstance(n.test, ast.Name):
                v = RepoParser._parse_variable(n.test)
                tokens.append(v)
            if hasattr(n.test, "comparators"):
                for item in n.test.comparators:
                    v = RepoParser._parse_variable(item)
                    if v:
                        tokens.append(v)
            return tokens
        except Exception as e:
            logger.warning(f"Unsupported if: {n}, err:{e}")
        return tokens

    @staticmethod
    def _parse_if_compare(n):
        """
        Parses an 'if' condition Abstract Syntax Tree (AST) node.

        Args:
            n: The AST node representing an 'if' condition.

        Returns:
            None or Parsed information from the 'if' condition node.
        """
        if hasattr(n, "left"):
            return RepoParser._parse_variable(n.left)
        else:
            return []

    @staticmethod
    def _parse_variable(node):
        """
        Parses a variable Abstract Syntax Tree (AST) node.

        Args:
            node: The AST node representing a variable.

        Returns:
            None or Parsed information from the variable node.
        """
        try:
            funcs = {
                any_to_str(ast.Constant): lambda x: x.value,
                any_to_str(ast.Name): lambda x: x.id,
                any_to_str(ast.Attribute): lambda x: f"{x.value.id}.{x.attr}"
                if hasattr(x.value, "id")
                else f"{x.attr}",
                any_to_str(ast.Call): lambda x: RepoParser._parse_variable(x.func),
                any_to_str(ast.Tuple): lambda x: [d.value for d in x.dims],
            }
            func = funcs.get(any_to_str(node))
            if not func:
                raise NotImplementedError(f"Not implement:{node}")
            return func(node)
        except Exception as e:
            logger.warning(f"Unsupported variable:{node}, err:{e}")

    @staticmethod
    def _parse_assign(node):
        """
        Parses an assignment Abstract Syntax Tree (AST) node.

        Args:
            node: The AST node representing an assignment.

        Returns:
            None or Parsed information from the assignment node.
        """
        return [RepoParser._parse_variable(t) for t in node.targets]

    async def rebuild_class_views(self, path: str | Path = None):
        """
        Executes `pylint` to reconstruct the dot format class view repository file.

        Args:
            path (str | Path): The path to the target directory or file. Default is None.
        """
        if not path:
            path = self.base_directory
        path = Path(path)
        if not path.exists():
            return
        init_file = path / "__init__.py"
        if not init_file.exists():
            raise ValueError("Failed to import module __init__ with error:No module named __init__.")
        command = f"pyreverse {str(path)} -o dot"
        output_dir = path / "__dot__"
        output_dir.mkdir(parents=True, exist_ok=True)
        result = subprocess.run(command, shell=True, check=True, cwd=str(output_dir))
        if result.returncode != 0:
            raise ValueError(f"{result}")
        class_view_pathname = output_dir / "classes.dot"
        class_views = await self._parse_classes(class_view_pathname)
        relationship_views = await self._parse_class_relationships(class_view_pathname)
        packages_pathname = output_dir / "packages.dot"
        class_views, relationship_views, package_root = RepoParser._repair_namespaces(
            class_views=class_views, relationship_views=relationship_views, path=path
        )
        class_view_pathname.unlink(missing_ok=True)
        packages_pathname.unlink(missing_ok=True)
        return class_views, relationship_views, package_root

    @staticmethod
    async def _parse_classes(class_view_pathname: Path) -> List[DotClassInfo]:
        """
        Parses a dot format class view repository file.

        Args:
            class_view_pathname (Path): The path to the dot format class view repository file.

        Returns:
            List[DotClassInfo]: A list of DotClassInfo objects representing the parsed classes.
        """
        class_views = []
        if not class_view_pathname.exists():
            return class_views
        data = await aread(filename=class_view_pathname, encoding="utf-8")
        lines = data.split("\n")
        for line in lines:
            package_name, info = RepoParser._split_class_line(line)
            if not package_name:
                continue
            class_name, members, functions = re.split(r"(?<!\\)\|", info)
            class_info = DotClassInfo(name=class_name)
            class_info.package = package_name
            for m in members.split("\n"):
                if not m:
                    continue
                attr = DotClassAttribute.parse(m)
                class_info.attributes[attr.name] = attr
                for i in attr.compositions:
                    if i not in class_info.compositions:
                        class_info.compositions.append(i)
            for f in functions.split("\n"):
                if not f:
                    continue
                method = DotClassMethod.parse(f)
                class_info.methods[method.name] = method
                for i in method.aggregations:
                    if i not in class_info.compositions and i not in class_info.aggregations:
                        class_info.aggregations.append(i)
            class_views.append(class_info)
        return class_views

    @staticmethod
    async def _parse_class_relationships(class_view_pathname: Path) -> List[DotClassRelationship]:
        """
        Parses a dot format class view repository file.

        Args:
            class_view_pathname (Path): The path to the dot format class view repository file.

        Returns:
            List[DotClassRelationship]: A list of DotClassRelationship objects representing the parsed class relationships.
        """
        relationship_views = []
        if not class_view_pathname.exists():
            return relationship_views
        data = await aread(filename=class_view_pathname, encoding="utf-8")
        lines = data.split("\n")
        for line in lines:
            relationship = RepoParser._split_relationship_line(line)
            if not relationship:
                continue
            relationship_views.append(relationship)
        return relationship_views

    @staticmethod
    def _split_class_line(line: str) -> (str, str):
        """
        Parses a dot format line about class info and returns the class name part and class members part.

        Args:
            line (str): The dot format line containing class information.

        Returns:
            Tuple[str, str]: A tuple containing the class name part and class members part.
        """
        part_splitor = '" ['
        if part_splitor not in line:
            return None, None
        ix = line.find(part_splitor)
        class_name = line[0:ix].replace('"', "")
        left = line[ix:]
        begin_flag = "label=<{"
        end_flag = "}>"
        if begin_flag not in left or end_flag not in left:
            return None, None
        bix = left.find(begin_flag)
        eix = left.rfind(end_flag)
        info = left[bix + len(begin_flag) : eix]
        info = re.sub(r"<br[^>]*>", "\n", info)
        return class_name, info

    @staticmethod
    def _split_relationship_line(line: str) -> DotClassRelationship:
        """
        Parses a dot format line about the relationship of two classes and returns 'Generalize', 'Composite',
        or 'Aggregate'.

        Args:
            line (str): The dot format line containing relationship information.

        Returns:
            DotClassRelationship: The object of relationship representing either 'Generalize', 'Composite',
            or 'Aggregate' relationship.
        """
        splitters = [" -> ", " [", "];"]
        idxs = []
        for tag in splitters:
            if tag not in line:
                return None
            idxs.append(line.find(tag))
        ret = DotClassRelationship()
        ret.src = line[0 : idxs[0]].strip('"')
        ret.dest = line[idxs[0] + len(splitters[0]) : idxs[1]].strip('"')
        properties = line[idxs[1] + len(splitters[1]) : idxs[2]].strip(" ")
        mappings = {
            'arrowhead="empty"': GENERALIZATION,
            'arrowhead="diamond"': COMPOSITION,
            'arrowhead="odiamond"': AGGREGATION,
        }
        for k, v in mappings.items():
            if k in properties:
                ret.relationship = v
                if v != GENERALIZATION:
                    ret.label = RepoParser._get_label(properties)
                break
        return ret

    @staticmethod
    def _get_label(line: str) -> str:
        """
        Parses a dot format line and returns the label information.

        Args:
            line (str): The dot format line containing label information.

        Returns:
            str: The label information parsed from the line.
        """
        tag = 'label="'
        if tag not in line:
            return ""
        ix = line.find(tag)
        eix = line.find('"', ix + len(tag))
        return line[ix + len(tag) : eix]

    @staticmethod
    def _create_path_mapping(path: str | Path) -> Dict[str, str]:
        """
        Creates a mapping table between source code files' paths and module names.

        Args:
            path (str | Path): The path to the source code files or directory.

        Returns:
            Dict[str, str]: A dictionary mapping source code file paths to their corresponding module names.
        """
        mappings = {
            str(path).replace("/", "."): str(path),
        }
        files = []
        try:
            directory_path = Path(path)
            if not directory_path.exists():
                return mappings
            for file_path in directory_path.iterdir():
                if file_path.is_file():
                    files.append(str(file_path))
                else:
                    subfolder_files = RepoParser._create_path_mapping(path=file_path)
                    mappings.update(subfolder_files)
        except Exception as e:
            logger.error(f"Error: {e}")
        for f in files:
            mappings[str(Path(f).with_suffix("")).replace("/", ".")] = str(f)

        return mappings

    @staticmethod
    def _repair_namespaces(
        class_views: List[DotClassInfo], relationship_views: List[DotClassRelationship], path: str | Path
    ) -> (List[DotClassInfo], List[DotClassRelationship], str):
        """
        Augments namespaces to the path-prefixed classes and relationships.

        Args:
            class_views (List[DotClassInfo]): List of DotClassInfo objects representing class views.
            relationship_views (List[DotClassRelationship]): List of DotClassRelationship objects representing
                relationships.
            path (str | Path): The path to the source code files or directory.

        Returns:
            Tuple[List[DotClassInfo], List[DotClassRelationship], str]: A tuple containing the augmented class views,
            relationships, and the root path of the package.
        """
        if not class_views:
            return [], [], ""
        c = class_views[0]
        full_key = str(path).lstrip("/").replace("/", ".")
        root_namespace = RepoParser._find_root(full_key, c.package)
        root_path = root_namespace.replace(".", "/")

        mappings = RepoParser._create_path_mapping(path=path)
        new_mappings = {}
        ix_root_namespace = len(root_namespace)
        ix_root_path = len(root_path)
        for k, v in mappings.items():
            nk = k[ix_root_namespace:]
            nv = v[ix_root_path:]
            new_mappings[nk] = nv

        for c in class_views:
            c.package = RepoParser._repair_ns(c.package, new_mappings)
        for _, v in enumerate(relationship_views):
            v.src = RepoParser._repair_ns(v.src, new_mappings)
            v.dest = RepoParser._repair_ns(v.dest, new_mappings)
        return class_views, relationship_views, str(path)[: len(root_path)]

    @staticmethod
    def _repair_ns(package: str, mappings: Dict[str, str]) -> str:
        """
        Replaces the package-prefix with the namespace-prefix.

        Args:
            package (str): The package to be repaired.
            mappings (Dict[str, str]): A dictionary mapping source code file paths to their corresponding packages.

        Returns:
            str: The repaired namespace.
        """
        file_ns = package
        ix = 0
        while file_ns != "":
            if file_ns not in mappings:
                ix = file_ns.rfind(".")
                file_ns = file_ns[0:ix]
                continue
            break
        if file_ns == "":
            return ""
        internal_ns = package[ix + 1 :]
        ns = mappings[file_ns] + ":" + internal_ns.replace(".", ":")
        return ns

    @staticmethod
    def _find_root(full_key: str, package: str) -> str:
        """
        Returns the package root path based on the key, which is the full path, and the package information.

        Args:
            full_key (str): The full key representing the full path.
            package (str): The package information.

        Returns:
            str: The package root path.
        """
        left = full_key
        while left != "":
            if left in package:
                break
            if "." not in left:
                break
            ix = left.find(".")
            left = left[ix + 1 :]
        ix = full_key.rfind(left)
        return "." + full_key[0:ix]


def is_func(node) -> bool:
    """
    Returns True if the given node represents a function.

    Args:
        node: The Abstract Syntax Tree (AST) node.

    Returns:
        bool: True if the node represents a function, False otherwise.
    """
    return isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))



import json
from typing import Optional

import chromadb
from pydantic import model_validator

from metagpt.actions import Action
from metagpt.const import DEFAULT_WORKSPACE_ROOT
from metagpt.environment.werewolf.const import RoleType
from metagpt.ext.werewolf.schema import RoleExperience
from metagpt.logs import logger
from metagpt.rag.engines.simple import SimpleEngine
from metagpt.rag.schema import ChromaIndexConfig, ChromaRetrieverConfig
from metagpt.utils.common import read_json_file, write_json_file

DEFAULT_COLLECTION_NAME = "role_reflection"  # FIXME: some hard code for now
PERSIST_PATH = DEFAULT_WORKSPACE_ROOT.joinpath("werewolf_game/chroma")
PERSIST_PATH.mkdir(parents=True, exist_ok=True)


class AddNewExperiences(Action):
    name: str = "AddNewExperience"
    collection_name: str = DEFAULT_COLLECTION_NAME
    delete_existing: bool = False
    engine: Optional[SimpleEngine] = None

    @model_validator(mode="after")
    def validate_collection(self):
        if self.engine:
            return
        if self.delete_existing:
            try:
                # implement engine `DELETE` method later
                chromadb.PersistentClient(PERSIST_PATH.as_posix()).delete_collection(self.collection_name)
            except Exception as exp:
                logger.error(f"delete chroma collection: {self.collection_name} failed, exp: {exp}")

        self.engine = SimpleEngine.from_objs(
            retriever_configs=[
                ChromaRetrieverConfig(
                    persist_path=PERSIST_PATH, collection_name=self.collection_name, metadata={"hnsw:space": "cosine"}
                )
            ]
        )

    def run(self, experiences: list[RoleExperience]):
        if not experiences:
            return
        for i, exp in enumerate(experiences):
            exp.id = f"{exp.profile}-{exp.name}-step{i}-round_{exp.round_id}"

        AddNewExperiences._record_experiences_local(experiences)

        self.engine.add_objs(experiences)

    def add_from_file(self, file_path):
        experiences = read_json_file(file_path)
        experiences = [RoleExperience.model_validate(item) for item in experiences]
        experiences = [exp for exp in experiences if len(exp.reflection) > 2]  # not "" or not '""'

        self.engine.add_objs(experiences)

    @staticmethod
    def _record_experiences_local(experiences: list[RoleExperience]):
        round_id = experiences[0].round_id
        version = experiences[0].version
        version = "test" if not version else version
        experiences = [exp.model_dump() for exp in experiences]

        experience_path = DEFAULT_WORKSPACE_ROOT.joinpath(f"werewolf_game/experiences/{version}")
        experience_path.mkdir(parents=True, exist_ok=True)
        save_path = f"{experience_path}/{round_id}.json"
        write_json_file(save_path, experiences)
        logger.info(f"experiences saved to {save_path}")


class RetrieveExperiences(Action):
    name: str = "RetrieveExperiences"
    collection_name: str = DEFAULT_COLLECTION_NAME
    has_experiences: bool = True
    engine: Optional[SimpleEngine] = None
    topk: int = 10

    @model_validator(mode="after")
    def validate_collection(self):
        if self.engine:
            return
        try:
            self.engine = SimpleEngine.from_index(
                index_config=ChromaIndexConfig(
                    persist_path=PERSIST_PATH, collection_name=self.collection_name, metadata={"hnsw:space": "cosine"}
                ),
                retriever_configs=[
                    ChromaRetrieverConfig(
                        similarity_top_k=self.topk,
                        persist_path=PERSIST_PATH,
                        collection_name=self.collection_name,
                        metadata={"hnsw:space": "cosine"},
                    )
                ],
            )
        except Exception as exp:
            logger.warning(f"No experience pool: {self.collection_name}, exp: {exp}")

    def run(self, query: str, profile: str, excluded_version: str = "", verbose: bool = False) -> str:
        """_summary_

        Args:
            query (str): 用当前的reflection作为query去检索过去相似的reflection
            profile (str): _description_

        Returns:
            _type_: _description_
        """
        if not self.engine or len(query) <= 2:  # not "" or not '""'
            logger.warning("engine is None or query too short")
            return ""

        # ablation experiment logic
        if profile == RoleType.WEREWOLF.value:  # role werewolf as baseline, don't use experiences
            logger.warning("Disable werewolves' experiences")
            return ""

        results = self.engine.retrieve(query)

        logger.info(f"retrieve {profile}'s experiences")
        experiences = [res.metadata["obj"] for res in results]

        past_experiences = []  # currently use post-process to filter, and later add `filters` in rag
        for exp in experiences:
            if exp.profile == profile and exp.version != excluded_version:
                past_experiences.append(exp)

        if verbose and results:
            logger.info("past_experiences: {}".format("\n\n".join(past_experiences)))
            distances = results[0].score
            logger.info(f"distances: {distances}")

        template = """
        {
            "Situation __i__": "__situation__"
            ,"Moderator's instruction": "__instruction__"
            ,"Your action or speech during that time": "__response__"
            ,"Reality": "In fact, it turned out the true roles are __game_step__",
            ,"Outcome": "You __outcome__ in the end"
        }
        """
        past_experiences = [
            (
                template.replace("__i__", str(i))
                .replace("__situation__", exp.reflection)
                .replace("__instruction__", exp.instruction)
                .replace("__response__", exp.response)
                .replace("__game_step__", exp.game_setup.replace("0 | Game setup:\n", "").replace("\n", " "))
                .replace("__outcome__", exp.outcome)
            )
            for i, exp in enumerate(past_experiences)
        ]
        logger.info("past_experiences: {}".format("\n".join(past_experiences)))
        logger.info("retrieval done")

        return json.dumps(past_experiences)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from sparkai.core.messages import _convert_to_message, convert_to_messages
from sparkai.core.messages.ai import AIMessage
from sparkai.core.messages.base import BaseMessage
from sparkai.core.messages.human import HumanMessage
from sparkai.core.messages.system import SystemMessage
from sparkai.core.outputs.llm_result import LLMResult
from sparkai.llm.llm import ChatSparkLLM

from metagpt.configs.llm_config import LLMConfig, LLMType
from metagpt.const import USE_CONFIG_TIMEOUT
from metagpt.logs import log_llm_stream
from metagpt.provider.base_llm import BaseLLM
from metagpt.provider.llm_provider_registry import register_provider
from metagpt.utils.common import any_to_str
from metagpt.utils.cost_manager import CostManager
from metagpt.utils.token_counter import SPARK_TOKENS


@register_provider(LLMType.SPARK)
class SparkLLM(BaseLLM):
    """
    用于讯飞星火大模型系列
    参考：https://github.com/iflytek/spark-ai-python"""

    def __init__(self, config: LLMConfig):
        self.config = config
        self.cost_manager = CostManager(token_costs=SPARK_TOKENS)
        self.model = self.config.domain
        self._init_client()

    def _init_client(self):
        self.client = ChatSparkLLM(
            spark_api_url=self.config.base_url,
            spark_app_id=self.config.app_id,
            spark_api_key=self.config.api_key,
            spark_api_secret=self.config.api_secret,
            spark_llm_domain=self.config.domain,
            streaming=True,
        )

    def _system_msg(self, msg: str) -> SystemMessage:
        return _convert_to_message(msg)

    def _user_msg(self, msg: str, **kwargs) -> HumanMessage:
        return _convert_to_message(msg)

    def _assistant_msg(self, msg: str) -> AIMessage:
        return _convert_to_message(msg)

    def get_choice_text(self, rsp: LLMResult) -> str:
        return rsp.generations[0][0].text

    def get_usage(self, response: LLMResult):
        message = response.generations[0][0].message
        if hasattr(message, "additional_kwargs"):
            return message.additional_kwargs.get("token_usage", {})
        else:
            return {}

    async def _achat_completion(self, messages: list[dict], timeout=USE_CONFIG_TIMEOUT):
        response = await self.acreate(messages, stream=False)
        usage = self.get_usage(response)
        self._update_costs(usage)
        return response

    async def acompletion(self, messages: list[dict], timeout=USE_CONFIG_TIMEOUT):
        return await self._achat_completion(messages, timeout)

    async def _achat_completion_stream(self, messages: list[dict], timeout: int = USE_CONFIG_TIMEOUT) -> str:
        response = await self.acreate(messages, stream=True)
        collected_content = []
        usage = {}
        async for chunk in response:
            collected_content.append(chunk.content)
            log_llm_stream(chunk.content)
            if hasattr(chunk, "additional_kwargs"):
                usage = chunk.additional_kwargs.get("token_usage", {})

        log_llm_stream("\n")
        self._update_costs(usage)
        full_content = "".join(collected_content)
        return full_content

    def _extract_assistant_rsp(self, context: list[BaseMessage]) -> str:
        return "\n".join([i.content for i in context if "AIMessage" in any_to_str(i)])

    async def acreate(self, messages: list[dict], stream: bool = True):
        messages = convert_to_messages(messages)
        if stream:
            return self.client.astream(messages)
        else:
            return await self.client.agenerate([messages])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : Stanford Town role

"""
Do the steps following:
- perceive, receive environment(Maze) info
- retrieve, retrieve memories
- plan, do plan like long-term plan and interact with Maze
- reflect, do the High-level thinking based on memories and re-add into the memory
- execute, move or else in the Maze
"""
import math
import random
import time
from datetime import datetime, timedelta
from operator import itemgetter
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from pydantic import ConfigDict, Field, field_validator, model_validator

from metagpt.actions.add_requirement import UserRequirement
from metagpt.environment.stanford_town.env_space import (
    EnvAction,
    EnvActionType,
    EnvObsParams,
    EnvObsType,
)
from metagpt.ext.stanford_town.actions.dummy_action import DummyAction, DummyMessage
from metagpt.ext.stanford_town.actions.inner_voice_action import (
    AgentWhisperThoughtAction,
)
from metagpt.ext.stanford_town.actions.run_reflect_action import AgentEventTriple
from metagpt.ext.stanford_town.memory.agent_memory import AgentMemory, BasicMemory
from metagpt.ext.stanford_town.memory.scratch import Scratch
from metagpt.ext.stanford_town.memory.spatial_memory import MemoryTree
from metagpt.ext.stanford_town.plan.st_plan import plan
from metagpt.ext.stanford_town.reflect.reflect import generate_poig_score, role_reflect
from metagpt.ext.stanford_town.utils.const import STORAGE_PATH, collision_block_id
from metagpt.ext.stanford_town.utils.mg_ga_transform import (
    get_role_environment,
    save_environment,
    save_movement,
)
from metagpt.ext.stanford_town.utils.utils import get_embedding, path_finder
from metagpt.logs import logger
from metagpt.roles.role import Role, RoleContext
from metagpt.schema import Message
from metagpt.utils.common import any_to_str

if TYPE_CHECKING:
    from metagpt.environment.stanford_town.stanford_town_env import (  # noqa: F401
        StanfordTownEnv,
    )


class STRoleContext(RoleContext):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    env: "StanfordTownEnv" = Field(default=None, exclude=True)
    memory: AgentMemory = Field(default_factory=AgentMemory)
    scratch: Scratch = Field(default_factory=Scratch)
    spatial_memory: MemoryTree = Field(default_factory=MemoryTree)

    @classmethod
    def model_rebuild(cls, **kwargs):
        from metagpt.environment.stanford_town.stanford_town_env import (  # noqa: F401
            StanfordTownEnv,
        )

        super(RoleContext, cls).model_rebuild(**kwargs)


class STRole(Role):
    # add a role's property structure to store role's age and so on like GA's Scratch.
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    name: str = Field(default="Klaus Mueller")
    profile: str = Field(default="STMember")

    rc: STRoleContext = Field(default_factory=STRoleContext)

    sim_code: str = Field(default="new_sim")
    step: int = Field(default=0)
    start_time: Optional[datetime] = Field(default=None)
    curr_time: Optional[datetime] = Field(default=None)
    sec_per_step: int = Field(default=10)
    game_obj_cleanup: dict = Field(default_factory=dict)
    inner_voice: bool = Field(default=False)
    has_inner_voice: bool = Field(default=False)

    role_storage_path: Optional[Path] = Field(default=None)

    @field_validator("curr_time", mode="before")
    @classmethod
    def check_curr_time(cls, curr_time: str) -> datetime:
        return datetime.strptime(curr_time, "%B %d, %Y, %H:%M:%S")

    @field_validator("start_time", mode="before")
    @classmethod
    def check_start_time(cls, start_time: str) -> datetime:
        return datetime.strptime(f"{start_time}, 00:00:00", "%B %d, %Y, %H:%M:%S")

    @model_validator(mode="after")
    def validate_st_role_after(self):
        self.role_storage_path = STORAGE_PATH.joinpath(f"{self.sim_code}/personas/{self.name}")

        self.load_from()  # load role's memory

        self.set_actions([])

        if self.has_inner_voice:
            # TODO add communication action
            self._watch([UserRequirement, DummyAction])
        else:
            self._watch([DummyAction])

    async def init_curr_tile(self):
        # init role
        role_env: dict = get_role_environment(self.sim_code, self.name, self.step)
        pt_x = role_env["x"]
        pt_y = role_env["y"]
        self.rc.scratch.curr_tile = (pt_x, pt_y)

        self.rc.env.step(
            EnvAction(
                action_type=EnvActionType.ADD_TILE_EVENT,
                coord=(pt_x, pt_y),
                event=self.scratch.get_curr_event_and_desc(),
            )
        )

    @property
    def scratch(self):
        return self.rc.scratch

    @property
    def role_tile(self):
        return self.scratch.curr_tile

    @property
    def a_mem(self):
        return self.rc.memory

    @property
    def s_mem(self):
        return self.rc.spatial_memory

    @property
    def memory(self):
        return self.rc.memory

    def load_from(self):
        """
        load role data from `storage/{simulation_name}/personas/{role_name}`
        """
        memory_saved = self.role_storage_path.joinpath("bootstrap_memory/associative_memory")
        self.rc.memory.set_mem_path(memory_saved)

        sp_mem_saved = self.role_storage_path.joinpath("bootstrap_memory/spatial_memory.json")
        self.rc.spatial_memory.set_mem_path(f_saved=sp_mem_saved)

        scratch_f_saved = self.role_storage_path.joinpath("bootstrap_memory/scratch.json")
        self.rc.scratch = Scratch.init_scratch_from_path(f_saved=scratch_f_saved)

        logger.info(f"Role: {self.name} loaded role's memory from {str(self.role_storage_path)}")

    def save_into(self):
        """
        save role data from `storage/{simulation_name}/personas/{role_name}`
        """
        memory_saved = self.role_storage_path.joinpath("bootstrap_memory/associative_memory")
        self.rc.memory.save(memory_saved)

        sp_mem_saved = self.role_storage_path.joinpath("bootstrap_memory/spatial_memory.json")
        self.rc.spatial_memory.save(sp_mem_saved)

        scratch_f_saved = self.role_storage_path.joinpath("bootstrap_memory/scratch.json")
        self.rc.scratch.save(scratch_f_saved)

        logger.info(f"Role: {self.name} saved role's memory into {str(self.role_storage_path)}")

    async def _observe(self, ignore_memory=False) -> int:
        if not self.rc.env:
            return 0
        news = []
        if not news:
            news = self.rc.msg_buffer.pop_all()
        old_messages = [] if ignore_memory else self.rc.memory.get()
        # Filter out messages of interest.
        self.rc.news = [
            n for n in news if (n.cause_by in self.rc.watch or self.name in n.send_to) and n not in old_messages
        ]

        if len(self.rc.news) == 1 and self.rc.news[0].cause_by == any_to_str(UserRequirement):
            logger.warning(f"Role: {self.name} add inner voice: {self.rc.news[0].content}")
            await self.add_inner_voice(self.rc.news[0].content)

        return 1  # always return 1 to execute role's `_react`

    async def add_inner_voice(self, whisper: str):
        async def generate_inner_thought(whisper: str):
            run_whisper_thought = AgentWhisperThoughtAction()
            inner_thought = await run_whisper_thought.run(self, whisper)
            return inner_thought

        thought = await generate_inner_thought(whisper)

        # init scratch curr_time with self.curr_time
        self.inner_voice = True
        self.rc.scratch.curr_time = self.curr_time

        created = self.rc.scratch.curr_time if self.rc.scratch.curr_time else datetime.now()
        expiration = created + timedelta(days=30)
        run_event_triple = AgentEventTriple()
        s, p, o = await run_event_triple.run(thought, self)
        keywords = set([s, p, o])
        thought_poignancy = await generate_poig_score(self, "event", whisper)
        thought_embedding_pair = (thought, get_embedding(thought))
        self.rc.memory.add_thought(
            created, expiration, s, p, o, thought, keywords, thought_poignancy, thought_embedding_pair, None
        )

    async def observe(self) -> list[BasicMemory]:
        # TODO observe info from maze_env
        """
        Perceive events around the role and saves it to the memory, both events
        and spaces.

        We first perceive the events nearby the role, as determined by its
        <vision_r>. If there are a lot of events happening within that radius, we
        take the <att_bandwidth> of the closest events. Finally, we check whether
        any of them are new, as determined by <retention>. If they are new, then we
        save those and return the <BasicMemory> instances for those events.

        OUTPUT:
            ret_events: a list of <BasicMemory> that are perceived and new.
        """
        # PERCEIVE SPACE
        # We get the nearby tiles given our current tile and the persona's vision
        # radius.
        nearby_tiles = self.rc.env.observe(
            EnvObsParams(
                obs_type=EnvObsType.TILE_NBR, coord=self.rc.scratch.curr_tile, vision_radius=self.rc.scratch.vision_r
            )
        )

        # We then store the perceived space. Note that the s_mem of the persona is
        # in the form of a tree constructed using dictionaries.
        for tile in nearby_tiles:
            tile_info = self.rc.env.observe(EnvObsParams(obs_type=EnvObsType.GET_TITLE, coord=tile))
            self.rc.spatial_memory.add_tile_info(tile_info)

        # PERCEIVE EVENTS.
        # We will perceive events that take place in the same arena as the
        # persona's current arena.

        curr_arena_path = self.rc.env.observe(
            EnvObsParams(obs_type=EnvObsType.TILE_PATH, coord=self.rc.scratch.curr_tile, level="arena")
        )

        # We do not perceive the same event twice (this can happen if an object is
        # extended across multiple tiles).
        percept_events_set = set()
        # We will order our percept based on the distance, with the closest ones
        # getting priorities.
        percept_events_list = []
        # First, we put all events that are occuring in the nearby tiles into the
        # percept_events_list
        for tile in nearby_tiles:
            tile_details = self.rc.env.observe(EnvObsParams(obs_type=EnvObsType.GET_TITLE, coord=tile))
            if tile_details["events"]:
                tmp_arena_path = self.rc.env.observe(
                    EnvObsParams(obs_type=EnvObsType.TILE_PATH, coord=tile, level="arena")
                )

                if tmp_arena_path == curr_arena_path:
                    # This calculates the distance between the persona's current tile,
                    # and the target tile.
                    dist = math.dist([tile[0], tile[1]], [self.rc.scratch.curr_tile[0], self.rc.scratch.curr_tile[1]])
                    # Add any relevant events to our temp set/list with the distant info.
                    for event in tile_details["events"]:
                        if event not in percept_events_set:
                            percept_events_list += [[dist, event]]
                            percept_events_set.add(event)

        # We sort, and perceive only self.rc.scratch.att_bandwidth of the closest
        # events. If the bandwidth is larger, then it means the persona can perceive
        # more elements within a small area.
        percept_events_list = sorted(percept_events_list, key=itemgetter(0))
        perceived_events = []
        for dist, event in percept_events_list[: self.rc.scratch.att_bandwidth]:
            perceived_events += [event]

        # Storing events.
        # <ret_events> is a list of <BasicMemory> instances from the persona's
        # associative memory.
        ret_events = []
        for p_event in perceived_events:
            s, p, o, desc = p_event
            if not p:
                # If the object is not present, then we default the event to "idle".
                p = "is"
                o = "idle"
                desc = "idle"
            desc = f"{s.split(':')[-1]} is {desc}"
            p_event = (s, p, o)

            # We retrieve the latest self.rc.scratch.retention events. If there is
            # something new that is happening (that is, p_event not in latest_events),
            # then we add that event to the a_mem and return it.
            latest_events = self.rc.memory.get_summarized_latest_events(self.rc.scratch.retention)
            if p_event not in latest_events:
                # We start by managing keywords.
                keywords = set()
                sub = p_event[0]
                obj = p_event[2]
                if ":" in p_event[0]:
                    sub = p_event[0].split(":")[-1]
                if ":" in p_event[2]:
                    obj = p_event[2].split(":")[-1]
                keywords.update([sub, obj])

                # Get event embedding
                desc_embedding_in = desc
                if "(" in desc:
                    desc_embedding_in = desc_embedding_in.split("(")[1].split(")")[0].strip()
                if desc_embedding_in in self.rc.memory.embeddings:
                    event_embedding = self.rc.memory.embeddings[desc_embedding_in]
                else:
                    event_embedding = get_embedding(desc_embedding_in)
                event_embedding_pair = (desc_embedding_in, event_embedding)

                # Get event poignancy.
                event_poignancy = await generate_poig_score(self, "event", desc_embedding_in)
                logger.debug(f"Role {self.name} event_poignancy: {event_poignancy}")

                # If we observe the persona's self chat, we include that in the memory
                # of the persona here.
                chat_node_ids = []
                if p_event[0] == f"{self.name}" and p_event[1] == "chat with":
                    curr_event = self.rc.scratch.act_event
                    if self.rc.scratch.act_description in self.rc.memory.embeddings:
                        chat_embedding = self.rc.memory.embeddings[self.rc.scratch.act_description]
                    else:
                        chat_embedding = get_embedding(self.rc.scratch.act_description)
                    chat_embedding_pair = (self.rc.scratch.act_description, chat_embedding)
                    chat_poignancy = await generate_poig_score(self, "chat", self.rc.scratch.act_description)
                    chat_node = self.rc.memory.add_chat(
                        self.rc.scratch.curr_time,
                        None,
                        curr_event[0],
                        curr_event[1],
                        curr_event[2],
                        self.rc.scratch.act_description,
                        keywords,
                        chat_poignancy,
                        chat_embedding_pair,
                        self.rc.scratch.chat,
                    )
                    chat_node_ids = [chat_node.memory_id]

                # Finally, we add the current event to the agent's memory.
                ret_events += [
                    self.rc.memory.add_event(
                        self.rc.scratch.curr_time,
                        None,
                        s,
                        p,
                        o,
                        desc,
                        keywords,
                        event_poignancy,
                        event_embedding_pair,
                        chat_node_ids,
                    )
                ]
                self.rc.scratch.importance_trigger_curr -= event_poignancy
                self.rc.scratch.importance_ele_n += 1

        return ret_events

    def retrieve(self, observed: list) -> dict:
        # TODO retrieve memories from agent_memory
        retrieved = dict()
        for event in observed:
            retrieved[event.description] = dict()
            retrieved[event.description]["curr_event"] = event

            relevant_events = self.rc.memory.retrieve_relevant_events(event.subject, event.predicate, event.object)
            retrieved[event.description]["events"] = list(relevant_events)

            relevant_thoughts = self.rc.memory.retrieve_relevant_thoughts(event.subject, event.predicate, event.object)
            retrieved[event.description]["thoughts"] = list(relevant_thoughts)

        return retrieved

    async def reflect(self):
        # TODO reflection if meet reflect condition
        await role_reflect(self)
        # TODO re-add result to memory
        # 已封装到Reflect函数之中

    async def execute(self, plan: str):
        """
        Args:
            plan: This is a string address of the action we need to execute.
            It comes in the form of "{world}:{sector}:{arena}:{game_objects}".
            It is important that you access this without doing negative
            indexing (e.g., [-1]) because the latter address elements may not be
            present in some cases.
            e.g., "dolores double studio:double studio:bedroom 1:bed"
        """
        roles = self.rc.env.get_roles()
        if "<random>" in plan and self.rc.scratch.planned_path == []:
            self.rc.scratch.act_path_set = False

        # <act_path_set> is set to True if the path is set for the current action.
        # It is False otherwise, and means we need to construct a new path.
        if not self.rc.scratch.act_path_set:
            # <target_tiles> is a list of tile coordinates where the persona may go
            # to execute the current action. The goal is to pick one of them.
            target_tiles = None
            logger.info(f"Role {self.name} plan: {plan}")

            if "<persona>" in plan:
                # Executing persona-persona interaction.
                target_p_tile = roles[plan.split("<persona>")[-1].strip()].scratch.curr_tile
                collision_maze = self.rc.env.observe()["collision_maze"]
                potential_path = path_finder(
                    collision_maze, self.rc.scratch.curr_tile, target_p_tile, collision_block_id
                )
                if len(potential_path) <= 2:
                    target_tiles = [potential_path[0]]
                else:
                    collision_maze = self.rc.env.observe()["collision_maze"]
                    potential_1 = path_finder(
                        collision_maze,
                        self.rc.scratch.curr_tile,
                        potential_path[int(len(potential_path) / 2)],
                        collision_block_id,
                    )

                    potential_2 = path_finder(
                        collision_maze,
                        self.rc.scratch.curr_tile,
                        potential_path[int(len(potential_path) / 2) + 1],
                        collision_block_id,
                    )
                    if len(potential_1) <= len(potential_2):
                        target_tiles = [potential_path[int(len(potential_path) / 2)]]
                    else:
                        target_tiles = [potential_path[int(len(potential_path) / 2 + 1)]]

            elif "<waiting>" in plan:
                # Executing interaction where the persona has decided to wait before
                # executing their action.
                x = int(plan.split()[1])
                y = int(plan.split()[2])
                target_tiles = [[x, y]]

            elif "<random>" in plan:
                # Executing a random location action.
                plan = ":".join(plan.split(":")[:-1])

                address_tiles = self.rc.env.observe()["address_tiles"]
                target_tiles = address_tiles[plan]
                target_tiles = random.sample(list(target_tiles), 1)

            else:
                # This is our default execution. We simply take the persona to the
                # location where the current action is taking place.
                # Retrieve the target addresses. Again, plan is an action address in its
                # string form. <maze.address_tiles> takes this and returns candidate
                # coordinates.
                address_tiles = self.rc.env.observe()["address_tiles"]
                if plan not in address_tiles:
                    address_tiles["Johnson Park:park:park garden"]  # ERRORRRRRRR
                else:
                    target_tiles = address_tiles[plan]

            # There are sometimes more than one tile returned from this (e.g., a tabe
            # may stretch many coordinates). So, we sample a few here. And from that
            # random sample, we will take the closest ones.
            if len(target_tiles) < 4:
                target_tiles = random.sample(list(target_tiles), len(target_tiles))
            else:
                target_tiles = random.sample(list(target_tiles), 4)
            # If possible, we want personas to occupy different tiles when they are
            # headed to the same location on the maze. It is ok if they end up on the
            # same time, but we try to lower that probability.
            # We take care of that overlap here.
            persona_name_set = set(roles.keys())
            new_target_tiles = []
            for i in target_tiles:
                access_tile = self.rc.env.observe(EnvObsParams(obs_type=EnvObsType.GET_TITLE, coord=i))
                curr_event_set = access_tile["events"]
                pass_curr_tile = False
                for j in curr_event_set:
                    if j[0] in persona_name_set:
                        pass_curr_tile = True
                if not pass_curr_tile:
                    new_target_tiles += [i]
            if len(new_target_tiles) == 0:
                new_target_tiles = target_tiles
            target_tiles = new_target_tiles

            # Now that we've identified the target tile, we find the shortest path to
            # one of the target tiles.
            curr_tile = self.rc.scratch.curr_tile
            closest_target_tile = None
            path = None
            for i in target_tiles:
                # path_finder takes a collision_mze and the curr_tile coordinate as
                # an input, and returns a list of coordinate tuples that becomes the
                # path.
                # e.g., [(0, 1), (1, 1), (1, 2), (1, 3), (1, 4)...]
                collision_maze = self.rc.env.observe()["collision_maze"]
                curr_path = path_finder(collision_maze, curr_tile, i, collision_block_id)
                if not closest_target_tile:
                    closest_target_tile = i
                    path = curr_path
                elif len(curr_path) < len(path):
                    closest_target_tile = i
                    path = curr_path

            # Actually setting the <planned_path> and <act_path_set>. We cut the
            # first element in the planned_path because it includes the curr_tile.
            self.rc.scratch.planned_path = path[1:]
            self.rc.scratch.act_path_set = True

        # Setting up the next immediate step. We stay at our curr_tile if there is
        # no <planned_path> left, but otherwise, we go to the next tile in the path.
        ret = self.rc.scratch.curr_tile
        if self.rc.scratch.planned_path:
            ret = self.rc.scratch.planned_path[0]
            self.rc.scratch.planned_path = self.rc.scratch.planned_path[1:]

        description = f"{self.rc.scratch.act_description}"
        description += f" @ {self.rc.scratch.act_address}"

        execution = ret, self.rc.scratch.act_pronunciatio, description
        return execution

    async def update_role_env(self) -> bool:
        role_env = get_role_environment(self.sim_code, self.name, self.step)
        ret = True
        if role_env:
            for key, val in self.game_obj_cleanup.items():
                self.rc.env.step(EnvAction(action_type=EnvActionType.TURN_TILE_EVENT_IDLE, coord=val, event=key))

            # reset game_obj_cleanup
            self.game_obj_cleanup = dict()
            curr_tile = self.role_tile
            new_tile = (role_env["x"], role_env["y"])
            self.rc.env.step(
                EnvAction(action_type=EnvActionType.RM_TITLE_SUB_EVENT, coord=curr_tile, subject=self.name)
            )
            self.rc.env.step(
                EnvAction(
                    action_type=EnvActionType.ADD_TILE_EVENT,
                    coord=new_tile,
                    event=self.scratch.get_curr_event_and_desc(),
                )
            )

            # the persona will travel to get to their destination. *Once*
            # the persona gets there, we activate the object action.
            if not self.scratch.planned_path:
                self.game_obj_cleanup[self.scratch.get_curr_event_and_desc()] = new_tile
                self.rc.env.step(
                    EnvAction(
                        action_type=EnvActionType.ADD_TILE_EVENT,
                        coord=new_tile,
                        event=self.scratch.get_curr_event_and_desc(),
                    )
                )

                blank = (self.scratch.get_curr_obj_event_and_desc()[0], None, None, None)
                self.rc.env.step(EnvAction(action_type=EnvActionType.RM_TILE_EVENT, coord=new_tile, event=blank))

            # update role's new tile
            self.rc.scratch.curr_tile = new_tile
        else:
            ret = False
            time.sleep(1)
            logger.warning(
                f"{self.sim_code}/environment/{self.step}.json not exist or parses failed, " f"sleep 1s and re-check"
            )
        return ret

    async def _react(self) -> Message:
        # update role env
        ret = await self.update_role_env()
        if not ret:
            # TODO add message
            logger.info(f"Role: {self.name} update_role_env return False")
            return DummyMessage()

        new_day = False
        if not self.scratch.curr_time or self.inner_voice:
            new_day = "First day"
        elif self.scratch.curr_time.strftime("%A %B %d") != self.curr_time.strftime("%A %B %d"):
            new_day = "New day"
        logger.info(f"Role: {self.name} new_day: {new_day}")
        self.rc.scratch.curr_time = self.curr_time

        # get maze_env from self.rc.env, and observe env info
        observed = await self.observe()

        # use self.rc.memory 's retrieve functions
        retrieved = self.retrieve(observed)

        plans = await plan(self, self.rc.env.get_roles(), new_day, retrieved)

        await self.reflect()

        # feed-back into maze_env
        next_tile, pronunciatio, description = await self.execute(plans)
        role_move = {
            "movement": next_tile,
            "pronunciatio": pronunciatio,
            "description": description,
            "chat": self.scratch.chat,
        }
        save_movement(self.name, role_move, step=self.step, sim_code=self.sim_code, curr_time=self.curr_time)

        # step update
        logger.info(f"Role: {self.name} run at {self.step} step on {self.curr_time} at tile: {self.scratch.curr_tile}")
        self.step += 1
        save_environment(self.name, self.step, self.sim_code, next_tile)
        self.curr_time += timedelta(seconds=self.sec_per_step)
        self.inner_voice = False

        time.sleep(0.5)
        return DummyMessage()


STRoleContext.model_rebuild()



"""Base Factory."""

from typing import Any, Callable


class GenericFactory:
    """Designed to get objects based on any keys."""

    def __init__(self, creators: dict[Any, Callable] = None):
        """Creators is a dictionary.

        Keys are identifiers, and the values are the associated creator function, which create objects.
        """
        self._creators = creators or {}

    def get_instances(self, keys: list[Any], **kwargs) -> list[Any]:
        """Get instances by keys."""
        return [self.get_instance(key, **kwargs) for key in keys]

    def get_instance(self, key: Any, **kwargs) -> Any:
        """Get instance by key.

        Raise Exception if key not found.
        """
        creator = self._creators.get(key)
        if creator:
            return creator(**kwargs)

        self._raise_for_key(key)

    def _raise_for_key(self, key: Any):
        raise ValueError(f"Creator not registered for key: {key}")


class ConfigBasedFactory(GenericFactory):
    """Designed to get objects based on object type."""

    def get_instance(self, key: Any, **kwargs) -> Any:
        """Get instance by the type of key.

        Key is config, such as a pydantic model, call func by the type of key, and the key will be passed to func.
        Raise Exception if key not found.
        """
        creator = self._creators.get(type(key))
        if creator:
            return creator(key, **kwargs)

        self._raise_for_key(key)

    def _raise_for_key(self, key: Any):
        raise ValueError(f"Unknown config: `{type(key)}`, {key}")

    @staticmethod
    def _val_from_config_or_kwargs(key: str, config: object = None, **kwargs) -> Any:
        """It prioritizes the configuration object's value unless it is None, in which case it looks into kwargs.

        Return None if not found.
        """
        if config is not None and hasattr(config, key):
            val = getattr(config, key)
            if val is not None:
                return val

        if key in kwargs:
            return kwargs[key]

        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Provider for volcengine.
See Also: https://console.volcengine.com/ark/region:ark+cn-beijing/model

config2.yaml example:
```yaml
llm:
  base_url: "https://ark.cn-beijing.volces.com/api/v3"
  api_type: "ark"
  endpoint: "ep-2024080514****-d****"
  api_key: "d47****b-****-****-****-d6e****0fd77"
  pricing_plan: "doubao-lite"
```
"""
from typing import Optional, Union

from pydantic import BaseModel
from volcenginesdkarkruntime import AsyncArk
from volcenginesdkarkruntime._base_client import AsyncHttpxClientWrapper
from volcenginesdkarkruntime._streaming import AsyncStream
from volcenginesdkarkruntime.types.chat import ChatCompletion, ChatCompletionChunk

from metagpt.configs.llm_config import LLMType
from metagpt.const import USE_CONFIG_TIMEOUT
from metagpt.logs import log_llm_stream
from metagpt.provider.llm_provider_registry import register_provider
from metagpt.provider.openai_api import OpenAILLM
from metagpt.utils.token_counter import DOUBAO_TOKEN_COSTS


@register_provider(LLMType.ARK)
class ArkLLM(OpenAILLM):
    """
    用于火山方舟的API
    见：https://www.volcengine.com/docs/82379/1263482
    """

    aclient: Optional[AsyncArk] = None

    def _init_client(self):
        """SDK: https://github.com/openai/openai-python#async-usage"""
        self.model = (
            self.config.endpoint or self.config.model
        )  # endpoint name, See more: https://console.volcengine.com/ark/region:ark+cn-beijing/endpoint
        self.pricing_plan = self.config.pricing_plan or self.model
        kwargs = self._make_client_kwargs()
        self.aclient = AsyncArk(**kwargs)

    def _make_client_kwargs(self) -> dict:
        kvs = {
            "ak": self.config.access_key,
            "sk": self.config.secret_key,
            "api_key": self.config.api_key,
            "base_url": self.config.base_url,
        }
        kwargs = {k: v for k, v in kvs.items() if v}

        # to use proxy, openai v1 needs http_client
        if proxy_params := self._get_proxy_params():
            kwargs["http_client"] = AsyncHttpxClientWrapper(**proxy_params)

        return kwargs

    def _update_costs(self, usage: Union[dict, BaseModel], model: str = None, local_calc_usage: bool = True):
        if next(iter(DOUBAO_TOKEN_COSTS)) not in self.cost_manager.token_costs:
            self.cost_manager.token_costs.update(DOUBAO_TOKEN_COSTS)
        if model in self.cost_manager.token_costs:
            self.pricing_plan = model
        if self.pricing_plan in self.cost_manager.token_costs:
            super()._update_costs(usage, self.pricing_plan, local_calc_usage)

    async def _achat_completion_stream(self, messages: list[dict], timeout=USE_CONFIG_TIMEOUT) -> str:
        response: AsyncStream[ChatCompletionChunk] = await self.aclient.chat.completions.create(
            **self._cons_kwargs(messages, timeout=self.get_timeout(timeout)),
            stream=True,
            extra_body={"stream_options": {"include_usage": True}},  # 只有增加这个参数才会在流式时最后返回usage
        )
        usage = None
        collected_messages = []
        async for chunk in response:
            chunk_message = chunk.choices[0].delta.content or "" if chunk.choices else ""  # extract the message
            log_llm_stream(chunk_message)
            collected_messages.append(chunk_message)
            if chunk.usage:
                # 火山方舟的流式调用会在最后一个chunk中返回usage,最后一个chunk的choices为[]
                usage = chunk.usage

        log_llm_stream("\n")
        full_reply_content = "".join(collected_messages)
        self._update_costs(usage, chunk.model)
        return full_reply_content

    async def _achat_completion(self, messages: list[dict], timeout=USE_CONFIG_TIMEOUT) -> ChatCompletion:
        kwargs = self._cons_kwargs(messages, timeout=self.get_timeout(timeout))
        rsp: ChatCompletion = await self.aclient.chat.completions.create(**kwargs)
        self._update_costs(rsp.usage, rsp.model)
        return rsp



"""Chroma retriever."""

from llama_index.core.retrievers import VectorIndexRetriever
from llama_index.core.schema import BaseNode


class ChromaRetriever(VectorIndexRetriever):
    """Chroma retriever."""

    def add_nodes(self, nodes: list[BaseNode], **kwargs) -> None:
        """Support add nodes."""
        self._index.insert_nodes(nodes, **kwargs)

    def persist(self, persist_dir: str, **kwargs) -> None:
        """Support persist.

        Chromadb automatically saves, so there is no need to implement."""



# -*- coding: utf-8 -*-
# @Time    : 2024/3/27 10:00
# @Author  : leiwu30
# @File    : stream_pipe.py
# @Version : None
# @Description : None

import json
import time
from multiprocessing import Pipe


class StreamPipe:
    def __init__(self, name=None):
        self.name = name
        self.parent_conn, self.child_conn = Pipe()
        self.finish: bool = False

    format_data = {
        "id": "chatcmpl-96bVnBOOyPFZZxEoTIGbdpFcVEnur",
        "object": "chat.completion.chunk",
        "created": 1711361191,
        "model": "gpt-3.5-turbo-0125",
        "system_fingerprint": "fp_3bc1b5746c",
        "choices": [
            {"index": 0, "delta": {"role": "assistant", "content": "content"}, "logprobs": None, "finish_reason": None}
        ],
    }

    def set_message(self, msg):
        self.parent_conn.send(msg)

    def get_message(self, timeout: int = 3):
        if self.child_conn.poll(timeout):
            return self.child_conn.recv()
        else:
            return None

    def msg2stream(self, msg):
        self.format_data["created"] = int(time.time())
        self.format_data["choices"][0]["delta"]["content"] = msg
        return f"data: {json.dumps(self.format_data, ensure_ascii=False)}\n".encode("utf-8")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   :

import re
from pathlib import Path
from typing import Union
from xml.etree.ElementTree import Element, iterparse

import cv2
import pyshine as ps

from metagpt.config2 import config
from metagpt.ext.android_assistant.utils.schema import (
    ActionOp,
    AndroidElement,
    BaseGridOpParam,
    BaseOpParam,
    Decision,
    GridOpParam,
    LongPressGridOpParam,
    LongPressOpParam,
    ReflectOp,
    RunState,
    SwipeGridOpParam,
    SwipeOpParam,
    TapGridOpParam,
    TapOpParam,
    TextOpParam,
)
from metagpt.logs import logger


def get_id_from_element(elem: Element) -> str:
    bounds = elem.attrib["bounds"][1:-1].split("][")
    x1, y1 = map(int, bounds[0].split(","))
    x2, y2 = map(int, bounds[1].split(","))
    elem_w, elem_h = x2 - x1, y2 - y1
    if "resource-id" in elem.attrib and elem.attrib["resource-id"]:
        elem_id = elem.attrib["resource-id"].replace(":", ".").replace("/", "_")
    else:
        elem_id = f"{elem.attrib['class']}_{elem_w}_{elem_h}"
    if "content-desc" in elem.attrib and elem.attrib["content-desc"] and len(elem.attrib["content-desc"]) < 20:
        content_desc = elem.attrib["content-desc"].replace("/", "_").replace(" ", "").replace(":", "_")
        elem_id += f"_{content_desc}"
    return elem_id


def traverse_xml_tree(xml_path: Path, elem_list: list[AndroidElement], attrib: str, add_index=False):
    path = []
    extra_config = config.extra
    for event, elem in iterparse(str(xml_path), ["start", "end"]):
        if event == "start":
            path.append(elem)
            if attrib in elem.attrib and elem.attrib[attrib] == "true":
                parent_prefix = ""
                if len(path) > 1:
                    parent_prefix = get_id_from_element(path[-2])
                bounds = elem.attrib["bounds"][1:-1].split("][")
                x1, y1 = map(int, bounds[0].split(","))
                x2, y2 = map(int, bounds[1].split(","))
                center = (x1 + x2) // 2, (y1 + y2) // 2
                elem_id = get_id_from_element(elem)
                if parent_prefix:
                    elem_id = parent_prefix + "_" + elem_id
                if add_index:
                    elem_id += f"_{elem.attrib['index']}"
                close = False
                for e in elem_list:
                    bbox = e.bbox
                    center_ = (bbox[0][0] + bbox[1][0]) // 2, (bbox[0][1] + bbox[1][1]) // 2
                    dist = (abs(center[0] - center_[0]) ** 2 + abs(center[1] - center_[1]) ** 2) ** 0.5
                    if dist <= extra_config.get("min_dist", 30):
                        close = True
                        break
                if not close:
                    elem_list.append(AndroidElement(uid=elem_id, bbox=((x1, y1), (x2, y2)), attrib=attrib))

        if event == "end":
            path.pop()


def elem_list_from_xml_tree(xml_path: Path, useless_list: list[str], min_dist: int) -> list[AndroidElement]:
    clickable_list = []
    focusable_list = []
    traverse_xml_tree(xml_path, clickable_list, "clickable", True)
    traverse_xml_tree(xml_path, focusable_list, "focusable", True)
    elem_list = []
    for elem in clickable_list:
        if elem.uid in useless_list:
            continue
        elem_list.append(elem)
    for elem in focusable_list:
        if elem.uid in useless_list:
            continue
        bbox = elem.bbox
        center = (bbox[0][0] + bbox[1][0]) // 2, (bbox[0][1] + bbox[1][1]) // 2
        close = False
        for e in clickable_list:
            bbox = e.bbox
            center_ = (bbox[0][0] + bbox[1][0]) // 2, (bbox[0][1] + bbox[1][1]) // 2
            dist = (abs(center[0] - center_[0]) ** 2 + abs(center[1] - center_[1]) ** 2) ** 0.5
            if dist <= min_dist:
                close = True
                break
        if not close:
            elem_list.append(elem)
    return elem_list


def draw_bbox_multi(
    img_path: Path,
    output_path: Path,
    elem_list: list[AndroidElement],
    record_mode: bool = False,
    dark_mode: bool = False,
):
    imgcv = cv2.imread(str(img_path))
    count = 1
    for elem in elem_list:
        try:
            top_left = elem.bbox[0]
            bottom_right = elem.bbox[1]
            left, top = top_left[0], top_left[1]
            right, bottom = bottom_right[0], bottom_right[1]
            label = str(count)
            if record_mode:
                if elem.attrib == "clickable":
                    color = (250, 0, 0)
                elif elem.attrib == "focusable":
                    color = (0, 0, 250)
                else:
                    color = (0, 250, 0)
                imgcv = ps.putBText(
                    imgcv,
                    label,
                    text_offset_x=(left + right) // 2 + 10,
                    text_offset_y=(top + bottom) // 2 + 10,
                    vspace=10,
                    hspace=10,
                    font_scale=1,
                    thickness=2,
                    background_RGB=color,
                    text_RGB=(255, 250, 250),
                    alpha=0.5,
                )
            else:
                text_color = (10, 10, 10) if dark_mode else (255, 250, 250)
                bg_color = (255, 250, 250) if dark_mode else (10, 10, 10)
                imgcv = ps.putBText(
                    imgcv,
                    label,
                    text_offset_x=(left + right) // 2 + 10,
                    text_offset_y=(top + bottom) // 2 + 10,
                    vspace=10,
                    hspace=10,
                    font_scale=1,
                    thickness=2,
                    background_RGB=bg_color,
                    text_RGB=text_color,
                    alpha=0.5,
                )
        except Exception as e:
            logger.error(f"ERROR: An exception occurs while labeling the image\n{e}")
        count += 1
    cv2.imwrite(str(output_path), imgcv)
    return imgcv


def draw_grid(img_path: Path, output_path: Path) -> tuple[int, int]:
    def get_unit_len(n):
        for i in range(1, n + 1):
            if n % i == 0 and 120 <= i <= 180:
                return i
        return -1

    image = cv2.imread(str(img_path))
    height, width, _ = image.shape
    color = (255, 116, 113)
    unit_height = get_unit_len(height)
    if unit_height < 0:
        unit_height = 120
    unit_width = get_unit_len(width)
    if unit_width < 0:
        unit_width = 120
    thick = int(unit_width // 50)
    rows = height // unit_height
    cols = width // unit_width
    for i in range(rows):
        for j in range(cols):
            label = i * cols + j + 1
            left = int(j * unit_width)
            top = int(i * unit_height)
            right = int((j + 1) * unit_width)
            bottom = int((i + 1) * unit_height)
            cv2.rectangle(image, (left, top), (right, bottom), color, thick // 2)
            cv2.putText(
                image,
                str(label),
                (left + int(unit_width * 0.05) + 3, top + int(unit_height * 0.3) + 3),
                0,
                int(0.01 * unit_width),
                (0, 0, 0),
                thick,
            )
            cv2.putText(
                image,
                str(label),
                (left + int(unit_width * 0.05), top + int(unit_height * 0.3)),
                0,
                int(0.01 * unit_width),
                color,
                thick,
            )
    cv2.imwrite(str(output_path), image)
    return rows, cols


def area_to_xy(area: int, subarea: str, width: int, height: int, rows: int, cols: int) -> tuple[int, int]:
    area -= 1
    row, col = area // cols, area % cols
    x_0, y_0 = col * (width // cols), row * (height // rows)
    if subarea == "top-left":
        x, y = x_0 + (width // cols) // 4, y_0 + (height // rows) // 4
    elif subarea == "top":
        x, y = x_0 + (width // cols) // 2, y_0 + (height // rows) // 4
    elif subarea == "top-right":
        x, y = x_0 + (width // cols) * 3 // 4, y_0 + (height // rows) // 4
    elif subarea == "left":
        x, y = x_0 + (width // cols) // 4, y_0 + (height // rows) // 2
    elif subarea == "right":
        x, y = x_0 + (width // cols) * 3 // 4, y_0 + (height // rows) // 2
    elif subarea == "bottom-left":
        x, y = x_0 + (width // cols) // 4, y_0 + (height // rows) * 3 // 4
    elif subarea == "bottom":
        x, y = x_0 + (width // cols) // 2, y_0 + (height // rows) * 3 // 4
    elif subarea == "bottom-right":
        x, y = x_0 + (width // cols) * 3 // 4, y_0 + (height // rows) * 3 // 4
    else:
        x, y = x_0 + (width // cols) // 2, y_0 + (height // rows) // 2
    return x, y


def elem_bbox_to_xy(bbox: tuple[tuple[int, int], tuple[int, int]]) -> tuple[int, int]:
    tl, br = bbox
    x, y = (tl[0] + br[0]) // 2, (tl[1] + br[1]) // 2
    return x, y


def reflect_parse_extarct(parsed_json: dict) -> ReflectOp:
    decision = parsed_json.get("Decision")
    if decision not in Decision.values():
        op = ReflectOp(param_state=RunState.FAIL)
    else:
        op = ReflectOp(
            decision=parsed_json.get("Decision"),
            thought=parsed_json.get("Thought"),
            documentation=parsed_json.get("Documentation"),
        )
    return op


def screenshot_parse_extract(
    parsed_json: dict, grid_on: bool = False
) -> Union[BaseOpParam, BaseGridOpParam, GridOpParam]:
    act = parsed_json.get("Action")
    last_act = parsed_json.get("Summary")
    act_name = act.split("(")[0]

    if RunState.FINISH.value.upper() in act:
        return BaseOpParam(param_state=RunState.FINISH)

    if grid_on:
        return screenshot_parse_extract_with_grid(act_name, act, last_act)
    else:
        return screenshot_parse_extract_without_grid(act_name, act, last_act)


def op_params_clean(params: list[str]) -> list[Union[int, str]]:
    param_values = []
    for param_value in params:
        if '"' in param_value or "'" in param_value:  # remove `"`
            param_values.append(param_value.strip()[1:-1])
        else:
            param_values.append(int(param_value))
    return param_values


def screenshot_parse_extract_without_grid(act_name: str, act: str, last_act: str) -> Union[BaseOpParam, GridOpParam]:
    if act_name == ActionOp.TAP.value:
        area = int(re.findall(r"tap\((.*?)\)", act)[0])
        op = TapOpParam(act_name=act_name, area=area, last_act=last_act)
    elif act_name == ActionOp.TEXT.value:
        input_str = re.findall(r"text\((.*?)\)", act)[0][1:-1]
        op = TextOpParam(act_name=act_name, input_str=input_str, last_act=last_act)
    elif act_name == ActionOp.LONG_PRESS.value:
        area = int(re.findall(r"long_press\((.*?)\)", act)[0])
        op = LongPressOpParam(act_name=act_name, area=area, last_act=last_act)
    elif act_name == ActionOp.SWIPE.value:
        params = re.findall(r"swipe\((.*?)\)", act)[0].split(",")
        params = op_params_clean(params)  # area, swipe_orient, dist
        op = SwipeOpParam(act_name=act_name, area=params[0], swipe_orient=params[1], dist=params[2], last_act=last_act)
    elif act_name == ActionOp.GRID.value:
        op = GridOpParam(act_name=act_name)
    else:
        op = BaseOpParam(param_state=RunState.FAIL)
    return op


def screenshot_parse_extract_with_grid(act_name: str, act: str, last_act: str) -> Union[BaseGridOpParam, GridOpParam]:
    if act_name == ActionOp.TAP.value:
        params = re.findall(r"tap\((.*?)\)", act)[0].split(",")
        params = op_params_clean(params)
        op = TapGridOpParam(act_name=act_name, area=params[0], subarea=params[1], last_act=last_act)
    elif act_name == ActionOp.LONG_PRESS.value:
        params = re.findall(r"long_press\((.*?)\)", act)[0].split(",")
        params = op_params_clean(params)
        op = LongPressGridOpParam(act_name=act_name, area=params[0], subarea=params[1], last_act=last_act)
    elif act_name == ActionOp.SWIPE.value:
        params = re.findall(r"swipe\((.*?)\)", act)[0].split(",")
        params = op_params_clean(params)
        op = SwipeGridOpParam(
            act_name=act_name, start_area=params[0], start_subarea=params[1], end_area=params[2], end_subarea=params[3]
        )
    elif act_name == ActionOp.GRID.value:
        op = GridOpParam(act_name=act_name)
    else:
        op = BaseGridOpParam(param_state=RunState.FAIL)
    return op



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2023/5/25 10:20
@Author  : alexanderwu
@File    : faiss_store.py
"""
import asyncio
from pathlib import Path
from typing import Any, Optional

import faiss
from llama_index.core import VectorStoreIndex, load_index_from_storage
from llama_index.core.embeddings import BaseEmbedding
from llama_index.core.schema import Document, QueryBundle, TextNode
from llama_index.core.storage import StorageContext
from llama_index.vector_stores.faiss import FaissVectorStore

from metagpt.document import IndexableDocument
from metagpt.document_store.base_store import LocalStore
from metagpt.logs import logger
from metagpt.utils.embedding import get_embedding


class FaissStore(LocalStore):
    def __init__(
        self, raw_data: Path, cache_dir=None, meta_col="source", content_col="output", embedding: BaseEmbedding = None
    ):
        self.meta_col = meta_col
        self.content_col = content_col
        self.embedding = embedding or get_embedding()
        self.store: VectorStoreIndex
        super().__init__(raw_data, cache_dir)

    def _load(self) -> Optional["VectorStoreIndex"]:
        index_file, store_file = self._get_index_and_store_fname()

        if not (index_file.exists() and store_file.exists()):
            logger.info("Missing at least one of index_file/store_file, load failed and return None")
            return None
        vector_store = FaissVectorStore.from_persist_dir(persist_dir=self.cache_dir)
        storage_context = StorageContext.from_defaults(persist_dir=self.cache_dir, vector_store=vector_store)
        index = load_index_from_storage(storage_context, embed_model=self.embedding)

        return index

    def _write(self, docs: list[str], metadatas: list[dict[str, Any]]) -> VectorStoreIndex:
        assert len(docs) == len(metadatas)
        documents = [Document(text=doc, metadata=metadatas[idx]) for idx, doc in enumerate(docs)]

        vector_store = FaissVectorStore(faiss_index=faiss.IndexFlatL2(1536))
        storage_context = StorageContext.from_defaults(vector_store=vector_store)
        index = VectorStoreIndex.from_documents(
            documents=documents, storage_context=storage_context, embed_model=self.embedding
        )

        return index

    def persist(self):
        self.store.storage_context.persist(self.cache_dir)

    def search(self, query: str, expand_cols=False, sep="\n", *args, k=5, **kwargs):
        retriever = self.store.as_retriever(similarity_top_k=k)
        rsp = retriever.retrieve(QueryBundle(query_str=query, embedding=self.embedding.get_text_embedding(query)))

        logger.debug(rsp)
        if expand_cols:
            return str(sep.join([f"{x.node.text}: {x.node.metadata}" for x in rsp]))
        else:
            return str(sep.join([f"{x.node.text}" for x in rsp]))

    async def asearch(self, *args, **kwargs):
        return await asyncio.to_thread(self.search, *args, **kwargs)

    def write(self):
        """Initialize the index and library based on the Document (JSON / XLSX, etc.) file provided by the user."""
        if not self.raw_data_path.exists():
            raise FileNotFoundError
        doc = IndexableDocument.from_path(self.raw_data_path, self.content_col, self.meta_col)
        docs, metadatas = doc.get_docs_and_metadatas()

        self.store = self._write(docs, metadatas)
        self.persist()
        return self.store

    def add(self, texts: list[str], *args, **kwargs) -> list[str]:
        """FIXME: Currently, the store is not updated after adding."""
        texts_embeds = self.embedding.get_text_embedding_batch(texts)
        nodes = [TextNode(text=texts[idx], embedding=embed) for idx, embed in enumerate(texts_embeds)]
        self.store.insert_nodes(nodes)

        return []

    def delete(self, *args, **kwargs):
        """Currently, faiss does not provide a delete interface."""
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : new_decomp_schedule

import datetime

from metagpt.ext.stanford_town.actions.st_action import STAction
from metagpt.logs import logger


class NewDecompSchedule(STAction):
    name: str = "NewDecompSchedule"

    def _func_validate(self, llm_resp: str, prompt: str) -> bool:
        resp = False
        try:
            llm_resp = self._func_cleanup(llm_resp, prompt)
            dur_sum = 0
            for act, dur in llm_resp:
                dur_sum += dur
                if isinstance(act, str):
                    return False
                if isinstance(dur, int):
                    return False
            x = prompt.split("\n")[0].split("originally planned schedule from")[-1].strip()[:-1]
            x = [datetime.datetime.strptime(i.strip(), "%H:%M %p") for i in x.split(" to ")]
            delta_min = int((x[1] - x[0]).total_seconds() / 60)

            if int(dur_sum) != int(delta_min):
                return False
        except Exception:
            pass
        return resp

    def _func_cleanup(self, llm_resp: str, prompt: str) -> list:
        new_schedule = prompt + " " + llm_resp.strip()
        new_schedule = new_schedule.split("The revised schedule:")[-1].strip()
        new_schedule = new_schedule.split("\n")

        ret_temp = []
        for i in new_schedule:
            ret_temp += [i.split(" -- ")]

        ret = []
        for time_str, action in ret_temp:
            start_time = time_str.split(" ~ ")[0].strip()
            end_time = time_str.split(" ~ ")[1].strip()
            delta = datetime.datetime.strptime(end_time, "%H:%M") - datetime.datetime.strptime(start_time, "%H:%M")
            delta_min = int(delta.total_seconds() / 60)
            if delta_min < 0:
                delta_min = 0
            ret += [[action, delta_min]]

        return ret

    def _func_fail_default_resp(self, main_act_dur: int, truncated_act_dur: int) -> int:
        dur_sum = 0
        for act, dur in main_act_dur:
            dur_sum += dur

        ret = truncated_act_dur[:]
        ret += main_act_dur[len(ret) - 1 :]

        # If there are access, we need to trim...
        ret_dur_sum = 0
        count = 0
        over = None
        for act, dur in ret:
            ret_dur_sum += dur
            if ret_dur_sum == dur_sum:
                break
            if ret_dur_sum > dur_sum:
                over = ret_dur_sum - dur_sum
                break
            count += 1

        if over:
            ret = ret[: count + 1]
            ret[-1][1] -= over

        return ret

    async def run(
        self,
        role: "STRole",
        main_act_dur: int,
        truncated_act_dur: int,
        start_time_hour: datetime,
        end_time_hour: datetime,
        inserted_act: str,
        inserted_act_dur: int,
        *args,
        **kwargs,
    ):
        def create_prompt_input(
            role: "STRole",
            main_act_dur: int,
            truncated_act_dur: int,
            start_time_hour: datetime,
            end_time_hour: datetime,
            inserted_act: str,
            inserted_act_dur: int,
        ):
            persona_name = role.name
            start_hour_str = start_time_hour.strftime("%H:%M %p")
            end_hour_str = end_time_hour.strftime("%H:%M %p")

            original_plan = ""
            for_time = start_time_hour
            for i in main_act_dur:
                original_plan += (
                    f'{for_time.strftime("%H:%M")} ~ '
                    f'{(for_time + datetime.timedelta(minutes=int(i[1]))).strftime("%H:%M")} -- ' + i[0]
                )
                original_plan += "\n"
                for_time += datetime.timedelta(minutes=int(i[1]))

            new_plan_init = ""
            for_time = start_time_hour
            for count, i in enumerate(truncated_act_dur):
                new_plan_init += (
                    f'{for_time.strftime("%H:%M")} ~ '
                    f'{(for_time + datetime.timedelta(minutes=int(i[1]))).strftime("%H:%M")} -- ' + i[0]
                )
                new_plan_init += "\n"
                if count < len(truncated_act_dur) - 1:
                    for_time += datetime.timedelta(minutes=int(i[1]))

            new_plan_init += (for_time + datetime.timedelta(minutes=int(i[1]))).strftime("%H:%M") + " ~"

            prompt_input = [
                persona_name,
                start_hour_str,
                end_hour_str,
                original_plan,
                persona_name,
                inserted_act,
                inserted_act_dur,
                persona_name,
                start_hour_str,
                end_hour_str,
                end_hour_str,
                new_plan_init,
            ]
            return prompt_input

        prompt_input = create_prompt_input(
            role, main_act_dur, truncated_act_dur, start_time_hour, end_time_hour, inserted_act, inserted_act_dur
        )
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, "new_decomp_schedule_v1.txt")
        self.fail_default_resp = self._func_fail_default_resp(main_act_dur, truncated_act_dur)
        output = await self._run_gpt35_max_tokens(prompt, max_tokens=1000)
        logger.info(f"Role: {role.name} Action: {self.cls_name} output: {output}")
        return output



# -*- encoding: utf-8 -*-
"""
@Date    :   2023/11/17 14:22:15
@Author  :   orange-crow
@File    :   execute_nb_code.py
"""
from __future__ import annotations

import asyncio
import base64
import re
from typing import Literal, Tuple

import nbformat
from nbclient import NotebookClient
from nbclient.exceptions import CellTimeoutError, DeadKernelError
from nbformat import NotebookNode
from nbformat.v4 import new_code_cell, new_markdown_cell, new_output
from rich.box import MINIMAL
from rich.console import Console, Group
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel
from rich.syntax import Syntax

from metagpt.actions import Action
from metagpt.logs import logger


class ExecuteNbCode(Action):
    """execute notebook code block, return result to llm, and display it."""

    nb: NotebookNode
    nb_client: NotebookClient
    console: Console
    interaction: str
    timeout: int = 600

    def __init__(
        self,
        nb=nbformat.v4.new_notebook(),
        timeout=600,
    ):
        super().__init__(
            nb=nb,
            nb_client=NotebookClient(nb, timeout=timeout),
            timeout=timeout,
            console=Console(),
            interaction=("ipython" if self.is_ipython() else "terminal"),
        )

    async def build(self):
        if self.nb_client.kc is None or not await self.nb_client.kc.is_alive():
            self.nb_client.create_kernel_manager()
            self.nb_client.start_new_kernel()
            self.nb_client.start_new_kernel_client()

    async def terminate(self):
        """kill NotebookClient"""
        if self.nb_client.km is not None and await self.nb_client.km.is_alive():
            await self.nb_client.km.shutdown_kernel(now=True)
            await self.nb_client.km.cleanup_resources()

            channels = [
                self.nb_client.kc.stdin_channel,  # The channel for handling standard input to the kernel.
                self.nb_client.kc.hb_channel,  # The channel for heartbeat communication between the kernel and client.
                self.nb_client.kc.control_channel,  # The channel for controlling the kernel.
            ]

            # Stops all the running channels for this kernel
            for channel in channels:
                if channel.is_alive():
                    channel.stop()

            self.nb_client.kc = None
            self.nb_client.km = None

    async def reset(self):
        """reset NotebookClient"""
        await self.terminate()

        # sleep 1s to wait for the kernel to be cleaned up completely
        await asyncio.sleep(1)
        await self.build()
        self.nb_client = NotebookClient(self.nb, timeout=self.timeout)

    def add_code_cell(self, code: str):
        self.nb.cells.append(new_code_cell(source=code))

    def add_markdown_cell(self, markdown: str):
        self.nb.cells.append(new_markdown_cell(source=markdown))

    def _display(self, code: str, language: Literal["python", "markdown"] = "python"):
        if language == "python":
            code = Syntax(code, "python", theme="paraiso-dark", line_numbers=True)
            self.console.print(code)
        elif language == "markdown":
            display_markdown(code)
        else:
            raise ValueError(f"Only support for python, markdown, but got {language}")

    def add_output_to_cell(self, cell: NotebookNode, output: str):
        """add outputs of code execution to notebook cell."""
        if "outputs" not in cell:
            cell["outputs"] = []
        else:
            cell["outputs"].append(new_output(output_type="stream", name="stdout", text=str(output)))

    def parse_outputs(self, outputs: list[str], keep_len: int = 2000) -> Tuple[bool, str]:
        """Parses the outputs received from notebook execution."""
        assert isinstance(outputs, list)
        parsed_output, is_success = [], True
        for i, output in enumerate(outputs):
            output_text = ""
            if output["output_type"] == "stream" and not any(
                tag in output["text"]
                for tag in ["| INFO     | metagpt", "| ERROR    | metagpt", "| WARNING  | metagpt", "DEBUG"]
            ):
                output_text = output["text"]
            elif output["output_type"] == "display_data":
                if "image/png" in output["data"]:
                    self.show_bytes_figure(output["data"]["image/png"], self.interaction)
                else:
                    logger.info(
                        f"{i}th output['data'] from nbclient outputs dont have image/png, continue next output ..."
                    )
            elif output["output_type"] == "execute_result":
                output_text = output["data"]["text/plain"]
            elif output["output_type"] == "error":
                output_text, is_success = "\n".join(output["traceback"]), False

            # handle coroutines that are not executed asynchronously
            if output_text.strip().startswith("<coroutine object"):
                output_text = "Executed code failed, you need use key word 'await' to run a async code."
                is_success = False

            output_text = remove_escape_and_color_codes(output_text)
            # The useful information of the exception is at the end,
            # the useful information of normal output is at the begining.
            output_text = output_text[:keep_len] if is_success else output_text[-keep_len:]

            parsed_output.append(output_text)
        return is_success, ",".join(parsed_output)

    def show_bytes_figure(self, image_base64: str, interaction_type: Literal["ipython", None]):
        image_bytes = base64.b64decode(image_base64)
        if interaction_type == "ipython":
            from IPython.display import Image, display

            display(Image(data=image_bytes))
        else:
            import io

            from PIL import Image

            image = Image.open(io.BytesIO(image_bytes))
            image.show()

    def is_ipython(self) -> bool:
        try:
            # 如果在Jupyter Notebook中运行，__file__ 变量不存在
            from IPython import get_ipython

            if get_ipython() is not None and "IPKernelApp" in get_ipython().config:
                return True
            else:
                return False
        except NameError:
            return False

    async def run_cell(self, cell: NotebookNode, cell_index: int) -> Tuple[bool, str]:
        """set timeout for run code.
        returns the success or failure of the cell execution, and an optional error message.
        """
        try:
            await self.nb_client.async_execute_cell(cell, cell_index)
            return self.parse_outputs(self.nb.cells[-1].outputs)
        except CellTimeoutError:
            assert self.nb_client.km is not None
            await self.nb_client.km.interrupt_kernel()
            await asyncio.sleep(1)
            error_msg = "Cell execution timed out: Execution exceeded the time limit and was stopped; consider optimizing your code for better performance."
            return False, error_msg
        except DeadKernelError:
            await self.reset()
            return False, "DeadKernelError"
        except Exception:
            return self.parse_outputs(self.nb.cells[-1].outputs)

    async def run(self, code: str, language: Literal["python", "markdown"] = "python") -> Tuple[str, bool]:
        """
        return the output of code execution, and a success indicator (bool) of code execution.
        """
        self._display(code, language)

        if language == "python":
            # add code to the notebook
            self.add_code_cell(code=code)

            # build code executor
            await self.build()

            # run code
            cell_index = len(self.nb.cells) - 1
            success, outputs = await self.run_cell(self.nb.cells[-1], cell_index)

            if "!pip" in code:
                success = False

            return outputs, success

        elif language == "markdown":
            # add markdown content to markdown cell in a notebook.
            self.add_markdown_cell(code)
            # return True, beacuse there is no execution failure for markdown cell.
            return code, True
        else:
            raise ValueError(f"Only support for language: python, markdown, but got {language}, ")


def remove_escape_and_color_codes(input_str: str):
    # 使用正则表达式去除jupyter notebook输出结果中的转义字符和颜色代码
    # Use regular expressions to get rid of escape characters and color codes in jupyter notebook output.
    pattern = re.compile(r"\x1b\[[0-9;]*[mK]")
    result = pattern.sub("", input_str)
    return result


def display_markdown(content: str):
    # Use regular expressions to match blocks of code one by one.
    matches = re.finditer(r"```(.+?)```", content, re.DOTALL)
    start_index = 0
    content_panels = []
    # Set the text background color and text color.
    style = "black on white"
    # Print the matching text and code one by one.
    for match in matches:
        text_content = content[start_index : match.start()].strip()
        code_content = match.group(0).strip()[3:-3]  # Remove triple backticks

        if text_content:
            content_panels.append(Panel(Markdown(text_content), style=style, box=MINIMAL))

        if code_content:
            content_panels.append(Panel(Markdown(f"```{code_content}"), style=style, box=MINIMAL))
        start_index = match.end()

    # Print remaining text (if any).
    remaining_text = content[start_index:].strip()
    if remaining_text:
        content_panels.append(Panel(Markdown(remaining_text), style=style, box=MINIMAL))

    # Display all panels in Live mode.
    with Live(auto_refresh=False, console=Console(), vertical_overflow="visible") as live:
        live.update(Group(*content_panels))
        live.refresh()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : summarize the content of agents' conversation

from metagpt.ext.stanford_town.actions.st_action import STAction
from metagpt.logs import logger


class SummarizeConv(STAction):
    name: str = "SummarizeConv"

    def _func_validate(self, llm_resp: str, prompt: str) -> bool:
        resp = False
        try:
            _ = self._func_cleanup(llm_resp, prompt)
            resp = True
        except Exception:
            pass
        return resp

    def _func_cleanup(self, llm_resp: str, prompt: str) -> str:
        ret = "conversing about " + llm_resp.strip()
        return ret

    def _func_fail_default_resp(self) -> str:
        return "conversing with a housemate about morning greetings"

    async def run(self, conv: list):
        def create_prompt_input(conversation: list):
            convo_str = ""
            for row in conversation:
                convo_str += f'{row[0]}: "{row[1]}"\n'
            prompt_input = [convo_str]
            return prompt_input

        prompt_input = create_prompt_input(conv)
        prompt = self.generate_prompt_with_tmpl_filename(prompt_input, "summarize_conversation_v1.txt")

        example_output = "conversing about what to eat for lunch"
        special_instruction = (
            "The output must continue the sentence above by filling in the <fill in> tag. "
            "Don't start with 'this is a conversation about...' Just finish the sentence "
            "but do not miss any important details (including who are chatting)."
        )
        output = await self._run_gpt35(prompt, example_output, special_instruction)
        logger.info(f"Action: {self.cls_name} output: {output}")
        return output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
models_config.py

This module defines the ModelsConfig class for handling configuration of LLM models.

Attributes:
    CONFIG_ROOT (Path): Root path for configuration files.
    METAGPT_ROOT (Path): Root path for MetaGPT files.

Classes:
    ModelsConfig (YamlModel): Configuration class for LLM models.
"""
from pathlib import Path
from typing import Dict, List, Optional

from pydantic import Field, field_validator

from metagpt.config2 import merge_dict
from metagpt.configs.llm_config import LLMConfig
from metagpt.const import CONFIG_ROOT, METAGPT_ROOT
from metagpt.utils.yaml_model import YamlModel


class ModelsConfig(YamlModel):
    """
    Configuration class for `models` in `config2.yaml`.

    Attributes:
        models (Dict[str, LLMConfig]): Dictionary mapping model names or types to LLMConfig objects.

    Methods:
        update_llm_model(cls, value): Validates and updates LLM model configurations.
        from_home(cls, path): Loads configuration from ~/.metagpt/config2.yaml.
        default(cls): Loads default configuration from predefined paths.
        get(self, name_or_type: str) -> Optional[LLMConfig]: Retrieves LLMConfig by name or API type.
    """

    models: Dict[str, LLMConfig] = Field(default_factory=dict)

    @field_validator("models", mode="before")
    @classmethod
    def update_llm_model(cls, value):
        """
        Validates and updates LLM model configurations.

        Args:
            value (Dict[str, Union[LLMConfig, dict]]): Dictionary of LLM configurations.

        Returns:
            Dict[str, Union[LLMConfig, dict]]: Updated dictionary of LLM configurations.
        """
        for key, config in value.items():
            if isinstance(config, LLMConfig):
                config.model = config.model or key
            elif isinstance(config, dict):
                config["model"] = config.get("model") or key
        return value

    @classmethod
    def from_home(cls, path):
        """
        Loads configuration from ~/.metagpt/config2.yaml.

        Args:
            path (str): Relative path to configuration file.

        Returns:
            Optional[ModelsConfig]: Loaded ModelsConfig object or None if file doesn't exist.
        """
        pathname = CONFIG_ROOT / path
        if not pathname.exists():
            return None
        return ModelsConfig.from_yaml_file(pathname)

    @classmethod
    def default(cls):
        """
        Loads default configuration from predefined paths.

        Returns:
            ModelsConfig: Default ModelsConfig object.
        """
        default_config_paths: List[Path] = [
            METAGPT_ROOT / "config/config2.yaml",
            CONFIG_ROOT / "config2.yaml",
        ]

        dicts = [ModelsConfig.read_yaml(path) for path in default_config_paths]
        final = merge_dict(dicts)
        return ModelsConfig(**final)

    def get(self, name_or_type: str) -> Optional[LLMConfig]:
        """
        Retrieves LLMConfig object by name or API type.

        Args:
            name_or_type (str): Name or API type of the LLM model.

        Returns:
            Optional[LLMConfig]: LLMConfig object if found, otherwise None.
        """
        if not name_or_type:
            return None
        model = self.models.get(name_or_type)
        if model:
            return model
        for m in self.models.values():
            if m.api_type == name_or_type:
                return m
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   :

import json
from http import HTTPStatus
from typing import Any, AsyncGenerator, Dict, List, Union

import dashscope
from dashscope.aigc.generation import Generation
from dashscope.api_entities.aiohttp_request import AioHttpRequest
from dashscope.api_entities.api_request_data import ApiRequestData
from dashscope.api_entities.api_request_factory import _get_protocol_params
from dashscope.api_entities.dashscope_response import (
    GenerationOutput,
    GenerationResponse,
    Message,
)
from dashscope.client.base_api import BaseAioApi
from dashscope.common.constants import SERVICE_API_PATH, ApiProtocol
from dashscope.common.error import (
    InputDataRequired,
    InputRequired,
    ModelRequired,
    UnsupportedApiProtocol,
)

from metagpt.const import USE_CONFIG_TIMEOUT
from metagpt.logs import log_llm_stream
from metagpt.provider.base_llm import BaseLLM, LLMConfig
from metagpt.provider.llm_provider_registry import LLMType, register_provider
from metagpt.utils.cost_manager import CostManager
from metagpt.utils.token_counter import DASHSCOPE_TOKEN_COSTS


def build_api_arequest(
    model: str, input: object, task_group: str, task: str, function: str, api_key: str, is_service=True, **kwargs
):
    (
        api_protocol,
        ws_stream_mode,
        is_binary_input,
        http_method,
        stream,
        async_request,
        query,
        headers,
        request_timeout,
        form,
        resources,
    ) = _get_protocol_params(kwargs)
    task_id = kwargs.pop("task_id", None)
    if api_protocol in [ApiProtocol.HTTP, ApiProtocol.HTTPS]:
        if not dashscope.base_http_api_url.endswith("/"):
            http_url = dashscope.base_http_api_url + "/"
        else:
            http_url = dashscope.base_http_api_url

        if is_service:
            http_url = http_url + SERVICE_API_PATH + "/"

        if task_group:
            http_url += "%s/" % task_group
        if task:
            http_url += "%s/" % task
        if function:
            http_url += function
        request = AioHttpRequest(
            url=http_url,
            api_key=api_key,
            http_method=http_method,
            stream=stream,
            async_request=async_request,
            query=query,
            timeout=request_timeout,
            task_id=task_id,
        )
    else:
        raise UnsupportedApiProtocol("Unsupported protocol: %s, support [http, https, websocket]" % api_protocol)

    if headers is not None:
        request.add_headers(headers=headers)

    if input is None and form is None:
        raise InputDataRequired("There is no input data and form data")

    request_data = ApiRequestData(
        model,
        task_group=task_group,
        task=task,
        function=function,
        input=input,
        form=form,
        is_binary_input=is_binary_input,
        api_protocol=api_protocol,
    )
    request_data.add_resources(resources)
    request_data.add_parameters(**kwargs)
    request.data = request_data
    return request


class AGeneration(Generation, BaseAioApi):
    @classmethod
    async def acall(
        cls,
        model: str,
        prompt: Any = None,
        history: list = None,
        api_key: str = None,
        messages: List[Message] = None,
        plugins: Union[str, Dict[str, Any]] = None,
        **kwargs,
    ) -> Union[GenerationResponse, AsyncGenerator[GenerationResponse, None]]:
        if (prompt is None or not prompt) and (messages is None or not messages):
            raise InputRequired("prompt or messages is required!")
        if model is None or not model:
            raise ModelRequired("Model is required!")
        task_group, function = "aigc", "generation"  # fixed value
        if plugins is not None:
            headers = kwargs.pop("headers", {})
            if isinstance(plugins, str):
                headers["X-DashScope-Plugin"] = plugins
            else:
                headers["X-DashScope-Plugin"] = json.dumps(plugins)
            kwargs["headers"] = headers
        input, parameters = cls._build_input_parameters(model, prompt, history, messages, **kwargs)

        api_key, model = BaseAioApi._validate_params(api_key, model)
        request = build_api_arequest(
            model=model,
            input=input,
            task_group=task_group,
            task=Generation.task,
            function=function,
            api_key=api_key,
            **kwargs,
        )
        response = await request.aio_call()
        is_stream = kwargs.get("stream", False)
        if is_stream:

            async def aresp_iterator(response):
                async for resp in response:
                    yield GenerationResponse.from_api_response(resp)

            return aresp_iterator(response)
        else:
            return GenerationResponse.from_api_response(response)


@register_provider(LLMType.DASHSCOPE)
class DashScopeLLM(BaseLLM):
    def __init__(self, llm_config: LLMConfig):
        self.config = llm_config
        self.use_system_prompt = False  # only some models support system_prompt
        self.__init_dashscope()
        self.cost_manager = CostManager(token_costs=self.token_costs)

    def __init_dashscope(self):
        self.model = self.config.model
        self.api_key = self.config.api_key
        self.token_costs = DASHSCOPE_TOKEN_COSTS
        self.aclient: AGeneration = AGeneration

        # check support system_message models
        support_system_models = [
            "qwen-",  # all support
            "llama2-",  # all support
            "baichuan2-7b-chat-v1",
            "chatglm3-6b",
        ]
        for support_model in support_system_models:
            if support_model in self.model:
                self.use_system_prompt = True

    def _const_kwargs(self, messages: list[dict], stream: bool = False) -> dict:
        kwargs = {
            "api_key": self.api_key,
            "model": self.model,
            "messages": messages,
            "stream": stream,
            "result_format": "message",
        }
        if self.config.temperature > 0:
            # different model has default temperature. only set when it"s specified.
            kwargs["temperature"] = self.config.temperature
        if stream:
            kwargs["incremental_output"] = True
        return kwargs

    def _check_response(self, resp: GenerationResponse):
        if resp.status_code != HTTPStatus.OK:
            raise RuntimeError(f"code: {resp.code}, request_id: {resp.request_id}, message: {resp.message}")

    def get_choice_text(self, output: GenerationOutput) -> str:
        return output.get("choices", [{}])[0].get("message", {}).get("content", "")

    def completion(self, messages: list[dict]) -> GenerationOutput:
        resp: GenerationResponse = self.aclient.call(**self._const_kwargs(messages, stream=False))
        self._check_response(resp)

        self._update_costs(dict(resp.usage))
        return resp.output

    async def _achat_completion(self, messages: list[dict], timeout: int = USE_CONFIG_TIMEOUT) -> GenerationOutput:
        resp: GenerationResponse = await self.aclient.acall(**self._const_kwargs(messages, stream=False))
        self._check_response(resp)
        self._update_costs(dict(resp.usage))
        return resp.output

    async def acompletion(self, messages: list[dict], timeout=USE_CONFIG_TIMEOUT) -> GenerationOutput:
        return await self._achat_completion(messages, timeout=self.get_timeout(timeout))

    async def _achat_completion_stream(self, messages: list[dict], timeout: int = USE_CONFIG_TIMEOUT) -> str:
        resp = await self.aclient.acall(**self._const_kwargs(messages, stream=True))
        collected_content = []
        usage = {}
        async for chunk in resp:
            self._check_response(chunk)
            content = chunk.output.choices[0]["message"]["content"]
            usage = dict(chunk.usage)  # each chunk has usage
            log_llm_stream(content)
            collected_content.append(content)
        log_llm_stream("\n")
        self._update_costs(usage)
        full_content = "".join(collected_content)
        return full_content



import asyncio
from typing import List, Tuple, Union

import evaluate
import jieba
from llama_index.core.embeddings import BaseEmbedding
from llama_index.core.evaluation import SemanticSimilarityEvaluator
from llama_index.core.schema import NodeWithScore
from pydantic import BaseModel

from metagpt.const import EXAMPLE_BENCHMARK_PATH
from metagpt.logs import logger
from metagpt.rag.factories import get_rag_embedding
from metagpt.utils.common import read_json_file


class DatasetInfo(BaseModel):
    name: str
    document_files: List[str]
    gt_info: List[dict]


class DatasetConfig(BaseModel):
    datasets: List[DatasetInfo]


class RAGBenchmark:
    def __init__(
        self,
        embed_model: BaseEmbedding = None,
    ):
        self.evaluator = SemanticSimilarityEvaluator(
            embed_model=embed_model or get_rag_embedding(),
        )

    def set_metrics(
        self,
        bleu_avg: float = 0.0,
        bleu_1: float = 0.0,
        bleu_2: float = 0.0,
        bleu_3: float = 0.0,
        bleu_4: float = 0.0,
        rouge_l: float = 0.0,
        semantic_similarity: float = 0.0,
        recall: float = 0.0,
        hit_rate: float = 0.0,
        mrr: float = 0.0,
        length: float = 0.0,
        generated_text: str = None,
        ground_truth_text: str = None,
        question: str = None,
    ):
        metrics = {
            "bleu-avg": bleu_avg,
            "bleu-1": bleu_1,
            "bleu-2": bleu_2,
            "bleu-3": bleu_3,
            "bleu-4": bleu_4,
            "rouge-L": rouge_l,
            "semantic similarity": semantic_similarity,
            "recall": recall,
            "hit_rate": hit_rate,
            "mrr": mrr,
            "length": length,
        }

        log = {
            "generated_text": generated_text,
            "ground_truth_text": ground_truth_text,
            "question": question,
        }

        return {"metrics": metrics, "log": log}

    def bleu_score(self, response: str, reference: str, with_penalty=False) -> Union[float, Tuple[float]]:
        f = lambda text: list(jieba.cut(text))
        bleu = evaluate.load(path="bleu")
        results = bleu.compute(predictions=[response], references=[[reference]], tokenizer=f)

        bleu_avg = results["bleu"]
        bleu1 = results["precisions"][0]
        bleu2 = results["precisions"][1]
        bleu3 = results["precisions"][2]
        bleu4 = results["precisions"][3]
        brevity_penalty = results["brevity_penalty"]

        if with_penalty:
            return bleu_avg, bleu1, bleu2, bleu3, bleu4
        else:
            return 0.0 if brevity_penalty == 0 else bleu_avg / brevity_penalty, bleu1, bleu2, bleu3, bleu4

    def rougel_score(self, response: str, reference: str) -> float:
        # pip install rouge_score
        f = lambda text: list(jieba.cut(text))
        rouge = evaluate.load(path="rouge")

        results = rouge.compute(predictions=[response], references=[[reference]], tokenizer=f, rouge_types=["rougeL"])
        score = results["rougeL"]
        return score

    def recall(self, nodes: list[NodeWithScore], reference_docs: list[str]) -> float:
        if nodes:
            total_recall = sum(any(node.text in doc for node in nodes) for doc in reference_docs)
            return total_recall / len(reference_docs)
        else:
            return 0.0

    def hit_rate(self, nodes: list[NodeWithScore], reference_docs: list[str]) -> float:
        if nodes:
            return 1.0 if any(node.text in doc for doc in reference_docs for node in nodes) else 0.0
        else:
            return 0.0

    def mean_reciprocal_rank(self, nodes: list[NodeWithScore], reference_docs: list[str]) -> float:
        mrr_sum = 0.0

        for i, node in enumerate(nodes, start=1):
            for doc in reference_docs:
                if text in doc:
                    mrr_sum += 1.0 / i
                    return mrr_sum

        return mrr_sum

    async def semantic_similarity(self, response: str, reference: str) -> float:
        result = await self.evaluator.aevaluate(
            response=response,
            reference=reference,
        )

        return result.score

    async def compute_metric(
        self,
        response: str = None,
        reference: str = None,
        nodes: list[NodeWithScore] = None,
        reference_doc: list[str] = None,
        question: str = None,
    ):
        recall = self.recall(nodes, reference_doc)
        bleu_avg, bleu1, bleu2, bleu3, bleu4 = self.bleu_score(response, reference)
        rouge_l = self.rougel_score(response, reference)
        hit_rate = self.hit_rate(nodes, reference_doc)
        mrr = self.mean_reciprocal_rank(nodes, reference_doc)

        similarity = await self.semantic_similarity(response, reference)

        result = self.set_metrics(
            bleu_avg,
            bleu1,
            bleu2,
            bleu3,
            bleu4,
            rouge_l,
            similarity,
            recall,
            hit_rate,
            mrr,
            len(response),
            response,
            reference,
            question,
        )

        return result

    @staticmethod
    def load_dataset(ds_names: list[str] = ["all"]):
        infos = read_json_file((EXAMPLE_BENCHMARK_PATH / "dataset_info.json").as_posix())
        dataset_config = DatasetConfig(
            datasets=[
                DatasetInfo(
                    name=name,
                    document_files=[
                        (EXAMPLE_BENCHMARK_PATH / name / file).as_posix() for file in info["document_file"]
                    ],
                    gt_info=read_json_file((EXAMPLE_BENCHMARK_PATH / name / info["gt_file"]).as_posix()),
                )
                for dataset_info in infos
                for name, info in dataset_info.items()
                if name in ds_names or "all" in ds_names
            ]
        )

        return dataset_config


if __name__ == "__main__":
    benchmark = RAGBenchmark()
    answer = "是的，根据提供的信息，2023年7月20日，应急管理部和财政部确实联合发布了《因灾倒塌、损坏住房恢复重建救助工作规范》的通知。这份《规范》旨在进一步规范因灾倒塌、损坏住房的恢复重建救助相关工作。它明确了地方各级政府负责实施救助工作，应急管理部和财政部则负责统筹指导。地方财政应安排足够的资金，中央财政也会提供适当的补助。救助资金将通过专账管理，并采取特定的管理方式。救助对象是那些因自然灾害导致住房倒塌或损坏，并向政府提出申请且符合条件的受灾家庭。相关部门将组织调查统计救助对象信息，并建立档案。此外，《规范》还强调了资金发放的具体方式和公开透明的要求。"
    ground_truth = "“启明行动”是为了防控儿童青少年的近视问题，并发布了《防控儿童青少年近视核心知识十条》。"
    bleu_avg, bleu1, bleu2, bleu3, bleu4 = benchmark.bleu_score(answer, ground_truth)
    rougeL_score = benchmark.rougel_score(answer, ground_truth)
    similarity = asyncio.run(benchmark.SemanticSimilarity(answer, ground_truth))

    logger.info(
        f"BLEU Scores: bleu_avg = {bleu_avg}, bleu1 = {bleu1}, bleu2 = {bleu2}, bleu3 = {bleu3}, bleu4 = {bleu4}, "
        f"RougeL Score: {rougeL_score}, "
        f"Semantic Similarity: {similarity}"
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2024/1/4 16:33
@Author  : alexanderwu
@File    : llm_config.py
"""
from enum import Enum
from typing import Optional

from pydantic import field_validator

from metagpt.const import CONFIG_ROOT, LLM_API_TIMEOUT, METAGPT_ROOT
from metagpt.utils.yaml_model import YamlModel


class LLMType(Enum):
    OPENAI = "openai"
    ANTHROPIC = "anthropic"
    CLAUDE = "claude"  # alias name of anthropic
    SPARK = "spark"
    ZHIPUAI = "zhipuai"
    FIREWORKS = "fireworks"
    OPEN_LLM = "open_llm"
    GEMINI = "gemini"
    METAGPT = "metagpt"
    AZURE = "azure"
    OLLAMA = "ollama"
    QIANFAN = "qianfan"  # Baidu BCE
    DASHSCOPE = "dashscope"  # Aliyun LingJi DashScope
    MOONSHOT = "moonshot"
    MISTRAL = "mistral"
    YI = "yi"  # lingyiwanwu
    OPENROUTER = "openrouter"
    BEDROCK = "bedrock"
    ARK = "ark"  # https://www.volcengine.com/docs/82379/1263482#python-sdk

    def __missing__(self, key):
        return self.OPENAI


class LLMConfig(YamlModel):
    """Config for LLM

    OpenAI: https://github.com/openai/openai-python/blob/main/src/openai/resources/chat/completions.py#L681
    Optional Fields in pydantic: https://docs.pydantic.dev/latest/migration/#required-optional-and-nullable-fields
    """

    api_key: str = "sk-"
    api_type: LLMType = LLMType.OPENAI
    base_url: str = "https://api.openai.com/v1"
    api_version: Optional[str] = None

    model: Optional[str] = None  # also stands for DEPLOYMENT_NAME
    pricing_plan: Optional[str] = None  # Cost Settlement Plan Parameters.

    # For Cloud Service Provider like Baidu/ Alibaba
    access_key: Optional[str] = None
    secret_key: Optional[str] = None
    endpoint: Optional[str] = None  # for self-deployed model on the cloud

    # For Spark(Xunfei), maybe remove later
    app_id: Optional[str] = None
    api_secret: Optional[str] = None
    domain: Optional[str] = None

    # For Chat Completion
    max_token: int = 4096
    temperature: float = 0.0
    top_p: float = 1.0
    top_k: int = 0
    repetition_penalty: float = 1.0
    stop: Optional[str] = None
    presence_penalty: float = 0.0
    frequency_penalty: float = 0.0
    best_of: Optional[int] = None
    n: Optional[int] = None
    stream: bool = True
    # https://cookbook.openai.com/examples/using_logprobs
    logprobs: Optional[bool] = None
    top_logprobs: Optional[int] = None
    timeout: int = 600

    # For Amazon Bedrock
    region_name: str = None

    # For Network
    proxy: Optional[str] = None

    # Cost Control
    calc_usage: bool = True

    @field_validator("api_key")
    @classmethod
    def check_llm_key(cls, v):
        if v in ["", None, "YOUR_API_KEY"]:
            repo_config_path = METAGPT_ROOT / "config/config2.yaml"
            root_config_path = CONFIG_ROOT / "config2.yaml"
            if root_config_path.exists():
                raise ValueError(
                    f"Please set your API key in {root_config_path}. If you also set your config in {repo_config_path}, \nthe former will overwrite the latter. This may cause unexpected result.\n"
                )
            elif repo_config_path.exists():
                raise ValueError(f"Please set your API key in {repo_config_path}")
            else:
                raise ValueError("Please set your API key in config2.yaml")
        return v

    @field_validator("timeout")
    @classmethod
    def check_timeout(cls, v):
        return v or LLM_API_TIMEOUT



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Desc   : the prompt templates of assistant learning and acting

screenshot_parse_template = """You are an agent that is trained to perform some basic tasks on a smartphone. You will be given a 
smartphone screenshot. The interactive UI elements on the screenshot are labeled with numeric tags starting from 1. The 
numeric tag of each interactive element is located in the center of the element.

You can call the following functions to control the smartphone:

1. tap(element: int)
This function is used to tap an UI element shown on the smartphone screen.
"element" is a numeric tag assigned to an UI element shown on the smartphone screen.
A simple use case can be tap(5), which taps the UI element labeled with the number 5.

2. text(text_input: str)
This function is used to insert text input in an input field/box. text_input is the string you want to insert and must 
be wrapped with double quotation marks. A simple use case can be text("Hello, world!"), which inserts the string 
"Hello, world!" into the input area on the smartphone screen. This function is usually callable when you see a keyboard 
showing in the lower half of the screen.

3. long_press(element: int)
This function is used to long press an UI element shown on the smartphone screen.
"element" is a numeric tag assigned to an UI element shown on the smartphone screen.
A simple use case can be long_press(5), which long presses the UI element labeled with the number 5.

4. swipe(element: int, direction: str, dist: str)
This function is used to swipe an UI element shown on the smartphone screen, usually a scroll view or a slide bar.
"element" is a numeric tag assigned to an UI element shown on the smartphone screen. "direction" is a string that 
represents one of the four directions: up, down, left, right. "direction" must be wrapped with double quotation 
marks. "dist" determines the distance of the swipe and can be one of the three options: short, medium, long. You should 
choose the appropriate distance option according to your need.
A simple use case can be swipe(21, "up", "medium"), which swipes up the UI element labeled with the number 21 for a 
medium distance.

5. grid()
You should call this function when you find the element you want to interact with is not labeled with a numeric tag and 
other elements with numeric tags cannot help with the task. The function will bring up a grid overlay to divide the 
smartphone screen into small areas and this will give you more freedom to choose any part of the screen to tap, long 
press, or swipe.
{ui_document}
The task you need to complete is to: {task_description}. Your past actions to proceed with this task are summarized as 
follows: {last_act}
Now, given the documentation and the following labeled screenshot, you need to think and call the function needed to 
proceed with the task. Your output should include three parts in the given format:

You can only take one action at a time, so please directly call the function."""

screenshot_parse_with_grid_template = """You are an agent that is trained to perform some basic tasks on a smartphone. You will be given 
a smartphone screenshot overlaid by a grid. The grid divides the screenshot into small square areas. Each area is 
labeled with an integer in the top-left corner.

You can call the following functions to control the smartphone:

1. tap(area: int, subarea: str)
This function is used to tap a grid area shown on the smartphone screen. "area" is the integer label assigned to a grid 
area shown on the smartphone screen. "subarea" is a string representing the exact location to tap within the grid area. 
It can take one of the nine values: center, top-left, top, top-right, left, right, bottom-left, bottom, and 
bottom-right.
A simple use case can be tap(5, "center"), which taps the exact center of the grid area labeled with the number 5.

2. long_press(area: int, subarea: str)
This function is used to long press a grid area shown on the smartphone screen. "area" is the integer label assigned to 
a grid area shown on the smartphone screen. "subarea" is a string representing the exact location to long press within 
the grid area. It can take one of the nine values: center, top-left, top, top-right, left, right, bottom-left, bottom, 
and bottom-right.
A simple use case can be long_press(7, "top-left"), which long presses the top left part of the grid area labeled with 
the number 7.

3. swipe(start_area: int, start_subarea: str, end_area: int, end_subarea: str)
This function is used to perform a swipe action on the smartphone screen, especially when you want to interact with a 
scroll view or a slide bar. "start_area" is the integer label assigned to the grid area which marks the starting 
location of the swipe. "start_subarea" is a string representing the exact location to begin the swipe within the grid 
area. "end_area" is the integer label assigned to the grid area which marks the ending location of the swipe. 
"end_subarea" is a string representing the exact location to end the swipe within the grid area.
The two subarea parameters can take one of the nine values: center, top-left, top, top-right, left, right, bottom-left, 
bottom, and bottom-right.
A simple use case can be swipe(21, "center", 25, "right"), which performs a swipe starting from the center of grid area 
21 to the right part of grid area 25.

The task you need to complete is to: {task_description}. Your past actions to proceed with this task are summarized as 
follows: {last_act}
Now, given the following labeled screenshot, you need to think and call the function needed to proceed with the task. 
Your output should include three parts in the given format:

You can only take one action at a time, so please directly call the function."""

screenshot_parse_self_explore_template = """You are an agent that is trained to complete certain tasks on a smartphone. You will be 
given a screenshot of a smartphone app. The interactive UI elements on the screenshot are labeled with numeric tags 
starting from 1. 

You can call the following functions to interact with those labeled elements to control the smartphone:

1. tap(element: int)
This function is used to tap an UI element shown on the smartphone screen.
"element" is a numeric tag assigned to an UI element shown on the smartphone screen.
A simple use case can be tap(5), which taps the UI element labeled with the number 5.

2. text(text_input: str)
This function is used to insert text input in an input field/box. text_input is the string you want to insert and must 
be wrapped with double quotation marks. A simple use case can be text("Hello, world!"), which inserts the string 
"Hello, world!" into the input area on the smartphone screen. This function is only callable when you see a keyboard 
showing in the lower half of the screen.

3. long_press(element: int)
This function is used to long press an UI element shown on the smartphone screen.
"element" is a numeric tag assigned to an UI element shown on the smartphone screen.
A simple use case can be long_press(5), which long presses the UI element labeled with the number 5.

4. swipe(element: int, direction: str, dist: str)
This function is used to swipe an UI element shown on the smartphone screen, usually a scroll view or a slide bar.
"element" is a numeric tag assigned to an UI element shown on the smartphone screen. "direction" is a string that 
represents one of the four directions: up, down, left, right. "direction" must be wrapped with double quotation 
marks. "dist" determines the distance of the swipe and can be one of the three options: short, medium, long. You should 
choose the appropriate distance option according to your need.
A simple use case can be swipe(21, "up", "medium"), which swipes up the UI element labeled with the number 21 for a 
medium distance.

The task you need to complete is to {task_description}. Your past actions to proceed with this task are summarized as 
follows: {last_act}
Now, given the following labeled screenshot, you need to think and call the function needed to proceed with the task. 
Your output should include three parts in the given format:

You can only take one action at a time, so please directly call the function."""

screenshot_parse_self_explore_reflect_template = """I will give you screenshots of a mobile app before and after {action} the UI 
element labeled with the number '{ui_element}' on the first screenshot. The numeric tag of each element is located at 
the center of the element. The action of {action} this UI element was described as follows:
{last_act}
The action was also an attempt to proceed with a larger task, which is to {task_desc}. Your job is to carefully analyze 
the difference between the two screenshots to determine if the action is in accord with the description above and at 
the same time effectively moved the task forward. Your output should be determined based on the following situations:
1. BACK
If you think the action navigated you to a page where you cannot proceed with the given task, you should go back to the 
previous interface. At the same time, describe the functionality of the UI element concisely in one or two sentences by 
observing the difference between the two screenshots. Notice that your description of the UI element should focus on 
the general function. Never include the numeric tag of the UI element in your description. You can use pronouns such as 
"the UI element" to refer to the element. Your output should be in the following format:
Decision: BACK
Thought: <explain why you think the last action is wrong and you should go back to the previous interface>
Documentation: <describe the function of the UI element>
2. INEFFECTIVE
If you find the action changed nothing on the screen (screenshots before and after the action are identical), you 
should continue to interact with other elements on the screen. Notice that if you find the location of the cursor 
changed between the two screenshots, then they are not identical. Your output should be in the following format:
Decision: INEFFECTIVE
Thought: <explain why you made this decision>
Documentation: <None>
3. CONTINUE
If you find the action changed something on the screen but does not reflect the action description above and did not 
move the given task forward, you should continue to interact with other elements on the screen. At the same time, 
describe the functionality of the UI element concisely in one or two sentences by observing the difference between the 
two screenshots. Notice that your description of the UI element should focus on the general function. Never include the 
numeric tag of the UI element in your description. You can use pronouns such as "the UI element" to refer to the 
element. Your output should be in the following format:
Decision: CONTINUE
Thought: <explain why you think the action does not reflect the action description above and did not move the given 
task forward>
Documentation: <describe the function of the UI element>
4. SUCCESS
If you think the action successfully moved the task forward (even though it did not completed the task), you should 
describe the functionality of the UI element concisely in one or two sentences. Notice that your description of the UI 
element should focus on the general function. Never include the numeric tag of the UI element in your description. You 
can use pronouns such as "the UI element" to refer to the element. Your output should be in the following format:
Decision: SUCCESS
Thought: <explain why you think the action successfully moved the task forward>
Documentation: <describe the function of the UI element>
"""
